import datetime
import re
from abc import abstractmethod, ABC

import hidateinfer
import hiurlparser
import jellyfish
import numpy as np
import pandas as pd
from metaphone import doublemetaphone

from optimus.helpers.constants import ProfilerDataTypes, ProfilerDataTypesNumeric
from optimus.helpers.core import one_list_to_val, one_tuple_to_val, val_to_list
from optimus.helpers.decorators import apply_to_categories
from optimus.helpers.logger import logger
from optimus.infer import is_list, is_list_of_list, is_valid_datetime_format, \
    is_list_of_int, is_list_of_str, \
    regex_int_compiled, regex_decimal_compiled, regex_credit_card_compiled, regex_email_compiled, \
    regex_phone_number_compiled, regex_http_code_compiled, \
    regex_gender_compiled, regex_social_security_number_compiled, regex_us_state_compiled, regex_zip_code_compiled, \
    regex_ipv4_address_compiled, regex_boolean_compiled, regex_full_url_compiled, \
    regex_BAN_compiled, \
    regex_uuid_compiled, regex_ipv6_address_compiled, regex_mac_address_compiled, regex_address_compiled, \
    regex_list_compiled, regex_dict_compiled, regex_tuple_compiled


# ^(?:(?P<protocol>[\w\d]+)(?:\:\/\/))?(?P<sub_domain>(?P<www>(?:www)?)(?:\.?)(?:(?:[\w\d-]+|\.)*?)?)(?:\.?)(?P<domain>[^./]+(?=\.))\.(?P<top_domain>com(?![^/|:?#]))?(?P<port>(:)(\d+))?(?P<path>(?P<dir>\/(?:[^/\r\n]+(?:/))+)?(?:\/?)(?P<file>[^?#\r\n]+)?)?(?:\#(?P<fragment>[^#?\r\n]*))?(?:\?(?P<query>.*(?=$)))*$


class BaseFunctions(ABC):
    """
    Functions for internal use or to be called using 'F': `from optimus.functions import F`
    Note: some methods needs to be static so they can be passed to a Dask worker.
    """

    _engine = None
    """The engine to be used for internal use:
        `pd` from `import pandas as pd` or `dask` from `import dask`
    """
    _partition_engine = None
    """(optional, defaults to `_engine`) The engine to be used for internal use
        in a distributed engine (if supported):
        `pd` from `import pandas as pd` which is used by `Dask`
    """

    def __init_subclass__(cls):

        if cls._partition_engine is None:
            cls._partition_engine = cls._engine

    def __init__(self, df=None):

        if self._engine is None:
            raise TypeError("Can't instantiate class without attribute '_engine'")

        if df is not None and getattr(df, "partitions", False):
            self.root = df
        else:
            self.root = None

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, name):
        if "__" not in name:
            type_msg = "" if self.root is None else f" using {type(self.root).__name__}"
            raise NotImplementedError(f"\"{name}\" is not available" + type_msg)

    @property
    def _functions(self):
        """
        Gets the set of functions available in the engine
        """
        return self._partition_engine

    @staticmethod
    def to_dict(_s):
        if hasattr(_s, "to_dict"):
            result = _s.to_dict()
        else:
            result = _s
        return result

    @staticmethod
    def to_list(_s):
        if hasattr(_s, "tolist"):
            return _s.tolist()
        else:
            return _s

    @staticmethod
    def to_list_one(_s):
        if hasattr(_s, "tolist"):
            return one_list_to_val(_s.tolist())
        else:
            return _s

    @property
    def n_partitions(self):
        if self.root is None:
            return 1
        return self.root.partitions()

    @staticmethod
    def delayed(func):
        """
        Convert function to a delayed function (Dummy method on unsupported engines)

        :param func:
        :return:
        """
        return func

    def from_delayed(self, delayed):
        """
        Convert delayed objects to a DataFrame or Series (Dummy method on unsupported engines)

        :param delayed:
        :return:
        """
        return delayed[0]

    def to_delayed(self, delayed):
        """
        Convert DataFrame or Series to a list of delayed objects (Dummy method on unsupported engines)

        :param delayed:
        :return:
        """
        return [delayed]

    def apply_delayed(self, series, func, *args, **kwargs):
        """
        Applies a function to a Series for every partition using apply

        :param series:
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        result = self.to_delayed(series)
        result = [partition.apply(func, *args, **kwargs) for partition in result]
        result = self.from_delayed(result)
        result.index = series.index
        return result

    def map_delayed(self, series, func, *args, **kwargs):
        """
        Applies a function to a Series for every partition using map

        :param series:
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        result = self.to_delayed(series)
        result = [partition.map(func, *args, **kwargs) for partition in result]
        result = self.from_delayed(result)
        result.index = series.index
        return result

    @property
    def constants(self):
        """

        :return:
        """
        from optimus.engines.base.constants import BaseConstants
        return BaseConstants()

    @staticmethod
    def sort_df(dfd, cols, ascending):
        """
        Sort rows taking into account one column

        :param dfd:
        :param cols:
        :param ascending:
        :return:
        """
        return dfd.sort_values(cols, ascending=ascending).reset_index(drop=True)

    @staticmethod
    def reverse_df(dfd):
        """
        Reverse rows
        """
        return dfd[::-1]

    @staticmethod
    def append(dfd, dfd2):
        """
        Append two dataframes
        """
        return dfd.append(dfd2)

    def _new_series(self, *args, **kwargs):
        """
        Creates a new series (also known as column)
        """
        return self._functions.Series(*args, **kwargs)

    def compute(self, *args, **kwargs):
        return one_tuple_to_val((*(a for a in args), *(kwargs[k] for k in kwargs)))

    @staticmethod
    def to_items(series) -> dict:
        """
        Convert series to a list of tuples [(index, value), ...]
        """
        df = series.reset_index()
        return df.to_dict(orient='split')['data']

    def _to_boolean(self, series):
        """
        Converts series to bool
        """
        return series.map(lambda v: bool(v), na_action=None).astype('bool')

    def to_boolean(self, series):
        return self._to_boolean(series)

    def _to_boolean_none(self, series):
        """
        Converts series to boolean
        """
        return series.map(lambda v: bool(v), na_action='ignore').astype('object')

    def to_boolean_none(self, series):
        return self._to_boolean_none(series)

    def _to_float(self, series):
        """
        Converts a series values to floats
        """
        return pd.to_numeric(series, errors='coerce', downcast='float')

    def to_float(self, series):
        """
        Converts a series values to floats
        """
        # print(series.dtype, "----",self.constants.NUMERIC_INTERNAL_TYPES)
        if not str(series.dtype) in self.constants.NUMERIC_INTERNAL_TYPES + self.constants.BOOLEAN_INTERNAL_TYPES:
            return self._to_float(series)
        else:
            return series

    def _to_integer(self, series, default=0):
        """
        Converts a series values to integers
        :param series:
        :param default:
        :return:
        """
        return pd.to_numeric(series, errors='coerce', downcast='int')

    def to_numeric(self, series, default=0):
        if not str(series.dtype) in self.constants.NUMERIC_INTERNAL_TYPES:
            return self._to_float(series)
        else:
            return series

    def to_integer(self, series, default=0):
        """
        Converts a series values to integers
        """
        return self._to_integer(series, default)

    def _to_datetime(self, series, format):
        """
        Converts a series values to datetimes
        """
        return series

    def to_datetime(self, series, format):
        """
        Converts a series values to datetimes
        """
        return self._to_datetime(series, format)

    def to_string(self, series):
        """
        Converts a series values to strings
        """
        if not str(series.dtype) in self.constants.STRING_INTERNAL_TYPES:
            return series.astype("string")
        else:
            return series

    def to_string_accessor(self, series):
        """
        Converts a series values to strings and returns it's main functions
        accessor
        """
        if not str(series.dtype) in self.constants.STRING_INTERNAL_TYPES:
            return series.astype(str).str
        else:
            return series.str

    @staticmethod
    def duplicated(dfd, keep, subset):
        """
        Mark the duplicated values in a DataFrame using a subset of columns
        """
        return dfd.duplicated(keep=keep, subset=subset)

    def impute(self, series, strategy, fill_value):
        """
        Impute missing values in a series
        """
        from sklearn.impute import SimpleImputer
        imputer = SimpleImputer(strategy=strategy, fill_value=fill_value)
        series_fit = series.dropna()
        if str(series.dtype) in self.constants.OBJECT_TYPES:
            series_fit = series_fit.astype(str)
        values = series_fit.values.reshape(-1, 1)
        if len(values):
            imputer.fit(values)
            return imputer.transform(series.fillna(np.nan).values.reshape(-1, 1))
        else:
            logger.warn("list to fit imputer is empty, try cols.fill_na instead.")
            return series

    # Aggregation
    def date_format(self, series):
        """
        Get the date format of a series
        """
        dtype = str(series.dtype)
        if dtype in self.constants.STRING_TYPES:
            series = series.astype(str)
            result = hidateinfer.infer(self.compute(series).values)
            if not is_valid_datetime_format(result) or True:
                result_series = self.date_formats(series)
                result = result_series.mode().head(1)[0]

                if not is_valid_datetime_format(result):
                    return False

            return result
        elif dtype in self.constants.DATETIME_INTERNAL_TYPES:
            return True

        return False

    def _series_dtype(self, series):
        return series.dtype

    def min(self, series, numeric=False, string=False):
        """
        Get the minimum value of a series
        """
        if numeric:
            series = self.to_float(series)

        if string or str(self._series_dtype(series)) in self.constants.STRING_TYPES:
            return self.to_string(series.dropna()).min()
        else:
            return series.min()

    def max(self, series, numeric=False, string=False):
        """
        Get the maximum value of a series
        """
        if numeric:
            series = self.to_float(series)

        if string or str(series.dtype) in self.constants.STRING_TYPES:
            return self.to_string(series.dropna()).max()
        else:
            return series.max()

    def mean(self, series):
        """
        Get the mean value of a series
        """
        return self.to_float(series).mean()

    def mode(self, series):
        """
        Get the modal value of a series
        """
        if str(series.dtype) not in self.constants.NUMERIC_INTERNAL_TYPES:
            return self.delayed(self.to_list_one)(self.to_string(series.dropna()).mode())
        else:
            return self.delayed(self.to_list_one)(series.mode())

    @staticmethod
    def crosstab(series, other):
        """
        Compute a cross tabulation of two series
        """
        return pd.crosstab(series, other)

    def std(self, series):
        """
        Get the standard deviation of a series
        """
        return self.to_float(series).std()

    def sum(self, series):
        """
        Get the sum of a series
        """
        return float(self.to_float(series).sum())

    def prod(self, series):
        """
        Get the sum of a series
        """
        return self.to_float(series).prod()

    def cumsum(self, series):
        """
        Get the cumulative sum of a series
        """
        return self.to_float(series).cumsum()

    def cumprod(self, series):
        """
        Get the cumulative product of a series
        """
        return self.to_float(series).cumprod()

    def cummax(self, series):
        """
        Get the cumulative maximum of a series
        """
        return self.to_float(series).cummax()

    def cummin(self, series):
        """
        Get the cumulative minimum of a series
        """
        return self.to_float(series).cummin()

    def var(self, series):
        """
        Get the variance of a series
        """
        return self.to_float(series).var()

    def count_uniques(self, series, estimate=False):
        """
        Get the count of unique values in a series
        """
        return self.to_string(series).nunique()

    def unique_values(self, series, estimate=False):
        """
        Get the unique values in a series
        """
        # Cudf can not handle null so we fill it with non zero values.
        return self.to_string(series).unique()

    @staticmethod
    def count_na(series):
        """
        Get the count of missing values in a series
        """
        return series.isna().sum()
        # return {"count_na": {col_name:  for col_name in columns}}
        # return np.count_nonzero(_df[_serie].isnull().values.ravel())
        # return cp.count_nonzero(_df[_serie].isnull().values.ravel())

    @staticmethod
    @abstractmethod
    def count_zeros(series, *args):
        pass

    @staticmethod
    @abstractmethod
    def kurtosis(series):
        pass

    @staticmethod
    @abstractmethod
    def skew(series):
        pass

    # import dask.dataframe as dd
    def mad(self, series, error=False, more=False, estimate=False):

        _series = self.to_float(series).dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            median_value = _series.quantile(0.5)
            mad_value = (_series - median_value).abs().quantile(0.5)
            if more:
                return {"mad": mad_value, "median": median_value}
            else:
                return mad_value

    # TODO: dask seems more efficient triggering multiple .min() task, one for every column
    # cudf seems to be calculate faster in on pass using df.min()
    def range(self, series):
        series = self.to_float(series)
        return {"min": series.min(), "max": series.max()}

    def percentile(self, series, values=0.5, error=False, estimate=False):

        _series = self.to_float(series).dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            return self.delayed(self.to_dict)(_series.quantile(values))

    # def radians(series):
    #     return series.to_float().radians()
    #
    # def degrees(series, *args):
    #     return call(series, method_name="degrees")

    ###########################

    def z_score(self, series):
        t = self.to_float(series)
        return t - t.mean() / t.std(ddof=0)

    def modified_z_score(self, series, estimate):
        series = self.to_float(series)

        _series = series.dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            median = _series.quantile(0.5)
            mad = (_series - median).abs().quantile(0.5)

            return abs(0.6745 * (series - median) / mad)

    def clip(self, series, lower_bound, upper_bound):
        return self.to_float(series).clip(float(lower_bound), float(upper_bound))

    @apply_to_categories
    def cut(self, series, bins, labels, default):
        if is_list_of_int(bins):
            return pd.cut(self.to_float(series), bins, include_lowest=True, labels=labels)
        elif is_list_of_str(bins):
            conditions = [series.str.contains(i) for i in bins]

            return np.select(conditions, labels, default=default)

    def qcut(self, series, quantiles):
        return pd.qcut(series, quantiles)

    def abs(self, series):
        return self.to_float(series).abs()

    def exp(self, series):
        return self.to_float(series).exp()

    def lemmatize_verbs(self, series):
        import nltk
        w_tokenizer = nltk.tokenize.WhitespaceTokenizer()
        lemmatizer = nltk.stem.WordNetLemmatizer()

        def lemmatize_verbs_map(text):
            return " ".join([lemmatizer.lemmatize(w, "v") for w in w_tokenizer.tokenize(text)])

        return self.map_delayed(self.to_string(series), lemmatize_verbs_map, na_action=None)

    def word_tokenize(self, series):
        import nltk
        w_tokenizer = nltk.tokenize.WhitespaceTokenizer()

        def lemmatize_verbs_map(text):
            return w_tokenizer.tokenize(text)

        return self.map_delayed(self.to_string(series), lemmatize_verbs_map, na_action=None, meta="object")

    @apply_to_categories
    def word_count(self, series):
        return self.word_tokenize(series).str.len()

    @apply_to_categories
    def len(self, series):
        return self.to_string_accessor(series).len()

    @staticmethod
    @abstractmethod
    def sqrt(series):
        pass

    def mod(self, series, other):
        return self.to_float(series).mod(other)

    def round(self, series, decimals):
        return self.to_float(series).round(decimals)

    def pow(self, series, exponent):
        return self.to_float(series).pow(exponent)

    def floor(self, series):
        return self.to_float(series).floor()

    # def trunc(self):
    #     series = self.series
    #     return series.to_float().truncate()

    @staticmethod
    @abstractmethod
    def radians(series):
        pass

    @staticmethod
    @abstractmethod
    def degrees(series):
        pass

    @staticmethod
    @abstractmethod
    def ln(series):
        pass

    @staticmethod
    @abstractmethod
    def log(series, base):
        pass

    @staticmethod
    @abstractmethod
    def ceil(series):
        pass

    @abstractmethod
    def sin(self, series):
        pass

    @abstractmethod
    def cos(self, series):
        pass

    @abstractmethod
    def tan(self, series):
        pass

    @abstractmethod
    def asin(self, series):
        pass

    @abstractmethod
    def acos(self, series):
        pass

    @abstractmethod
    def atan(self, series):
        pass

    @abstractmethod
    def sinh(self, series):
        pass

    @abstractmethod
    def cosh(self, series):
        pass

    @abstractmethod
    def tanh(self, series):
        pass

    @abstractmethod
    def asinh(self, series):
        pass

    @abstractmethod
    def acosh(self, series):
        pass

    @abstractmethod
    def atanh(self, series):
        pass

    # Strings
    @apply_to_categories
    def match(self, series, regex):
        return self.to_string_accessor(series).match(regex)

    @apply_to_categories
    def lower(self, series):
        return self.to_string_accessor(series).lower()

    @apply_to_categories
    def upper(self, series):
        return self.to_string_accessor(series).upper()

    @apply_to_categories
    def title(self, series):
        return self.to_string_accessor(series).title()

    @apply_to_categories
    def capitalize(self, series):
        return self.to_string_accessor(series).capitalize()

    @apply_to_categories
    def pad(self, series, width, side, fillchar=""):
        return self.to_string_accessor(series).pad(width, side, fillchar)

    @apply_to_categories
    def extract(self, series, regex):
        return self.to_string_accessor(series).extract(regex)

    @apply_to_categories
    def slice(self, series, start, stop, step):
        return self.to_string_accessor(series).slice(start, stop, step)

    @apply_to_categories
    def trim(self, series):
        return self.to_string_accessor(series).strip()

    @apply_to_categories
    def strip(self, series, chars=None, side="both"):
        if side == "left":
            return self.to_string_accessor(series).lstrip(to_strip=chars)
        elif side == "right":
            return self.to_string_accessor(series).rstrip(to_strip=chars)
        else:
            return self.to_string_accessor(series).strip(to_strip=chars)

    @apply_to_categories
    def strip_html(self, series):
        return self.to_string(series).replace('<.*?>', '', regex=True)

    def _replace_string(self, series, to_replace, value, regex):
        return series.replace(to_replace, value, regex=regex)

    def replace_chars(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)%s' % re.escape(s) for s in search]
        else:
            str_regex = [r'%s' % re.escape(s) for s in search]
        return self._replace_string(self.to_string(series), str_regex, replace_by, regex=True)

    def replace_words(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)\b%s\b' % re.escape(s) for s in search]
        else:
            str_regex = [r'\b%s\b' % re.escape(s) for s in search]
        return self._replace_string(self.to_string(series), str_regex, replace_by, regex=True)

    def replace_full(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            regex = True
            str_search = [r'(?i)^%s$' % re.escape(s) for s in search]
        else:
            regex = False
            str_search = search
        return series.replace(str_search, replace_by, regex=regex)

    def replace_values(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)

        if ignore_case:
            regex = True
            search = [(r'(?i)^%s$' % re.escape(s)) for s in search]
        else:
            regex = False

        if is_list(replace_by) and is_list_of_list(search):
            for _s, _r in zip(search, replace_by):
                series = series.replace(_s, _r, regex=regex)

        else:
            series = series.replace(search, replace_by, regex=regex)

        return series

    def replace_regex_chars(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)%s' % s for s in search]
        else:
            str_regex = [r'%s' % s for s in search]
        return self._replace_string(self.to_string(series), str_regex, replace_by, regex=True)

    def replace_regex_words(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)\b%s\b' % s for s in search]
        else:
            str_regex = [r'\b%s\b' % s for s in search]
        return self._replace_string(self.to_string(series), str_regex, replace_by, regex=True)

    def replace_regex_full(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)^%s$' % s for s in search]
        else:
            str_regex = [r'^%s$' % s for s in search]
        return series.replace(str_regex, replace_by, regex=True)

    @apply_to_categories
    def remove_numbers(self, series):
        return self.to_string_accessor(series).replace(r'\d+', '', regex=True)

    @apply_to_categories
    def remove_white_spaces(self, series):
        return self.to_string_accessor(series).replace(" ", "")

    @apply_to_categories
    def remove_urls(self, series):
        return self.to_string_accessor(series).replace(r"https?://\S+|www\.\S+", "", regex=True)

    @apply_to_categories
    def normalize_spaces(self, series):
        return self.to_string_accessor(series).replace(r" +", " ", regex=True)

    def expand_contractions(self, series):
        pass

    # @staticmethod
    # def len(series):
    #     return series.str.len()

    def normalize_chars(self, series):
        pass

    @apply_to_categories
    def find(self, sub, start=0, end=None):
        series = self.series
        return self.to_string_accessor(series).find(sub, start, end)

    @apply_to_categories
    def rfind(self, series, sub, start=0, end=None):
        return self.to_string_accessor(series).rfind(sub, start, end)

    @apply_to_categories
    def left(self, series, position):
        return self.to_string_accessor(series)[:position]

    @apply_to_categories
    def right(self, series, position):
        return self.to_string_accessor(series)[-1 * position:]

    @apply_to_categories
    def mid(self, series, start, end):
        return self.to_string_accessor(series)[start:end]

    @apply_to_categories
    def starts_with(self, series, pat):
        return self.to_string_accessor(series).startswith(pat)

    @apply_to_categories
    def ends_with(self, series, pat):
        return self.to_string_accessor(series).endswith(pat)

    @apply_to_categories
    def contains(self, series, value, case, flags, na, regex):
        return self.to_string_accessor(series).contains(value, case=case, flags=flags, na=na, regex=regex)

    @apply_to_categories
    def char(self, series, _n):
        return self.to_string_accessor(series)[_n]

    @staticmethod
    def unicode(series):
        pass

    @staticmethod
    def exact(series, pat):
        return series == pat

    # dates
    def year(self, series, format):
        """
        Extract the year from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.year

    def month(self, series, format):
        """
        Extract the month from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.month

    def day(self, series, format):
        """
        Extract the day from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.day

    def hour(self, series, format):
        """
        Extract the hour from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.hour

    def minute(self, series, format):
        """
        Extract the minute from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.minute

    def second(self, series, format):
        """
        Extract the second from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.second

    def weekday(self, series, format):
        """
        Extract the weekday from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.weekday

    @staticmethod
    @abstractmethod
    def format_date(self, current_format=None, output_format=None):
        pass

    @staticmethod
    @abstractmethod
    def time_between(self, value=None, date_format=None):
        pass

    def years_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days / 365.25

    def months_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days / 30.436875

    def days_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days

    def hours_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 24.0 + series.dt.seconds / 3600.0

    def minutes_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 1440.0 + series.dt.seconds / 60.0

    def seconds_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 86400 + series.dt.seconds

    def domain(self, series):

        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["domain"], na_action=None)

    def top_domain(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["top_domain"], na_action=None)

    def sub_domain(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["sub_domain"], na_action=None)

    def url_scheme(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["protocol"], na_action=None)

    def url_path(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["path"], na_action=None)

    def url_file(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["file"], na_action=None)

    def url_query(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["query"], na_action=None)

    def url_fragment(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["fragment"], na_action=None)

    def host(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["host"], na_action=None)

    def port(self, series):
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["port"], na_action=None)

    @apply_to_categories
    def email_username(self, series):
        return self.to_string_accessor(series).split('@').str[0]

    @apply_to_categories
    def email_domain(self, series):
        return self.to_string_accessor(series).split('@').str[1]

    def infer_data_types(self, series):
        PDTN = ProfilerDataTypesNumeric
        default_value = 127

        if str(series.dtype) in self.constants.INT_INTERNAL_TYPES:
            series_result = PDTN.INT.value
        elif str(series.dtype) in self.constants.NUMERIC_INTERNAL_TYPES:
            series_result = pd.Series([0] * len(series), dtype=np.uint8)
            series_result[series.isna()] = PDTN.NULL.value
            mask = series % 1 == 0
            series_result[mask | ~series.isna()] = PDTN.INT.value
            series_result[~mask & ~series.isna()] = PDTN.FLOAT.value
        elif str(series.dtype) in self.constants.BOOLEAN_INTERNAL_TYPES:
            series_result = PDTN.BOOLEAN.value
        # elif str(series.dtype) in self.constants.CATEGORY_INTERNAL_TYPES:
        #     series_result = PDTN.CATEGORICAL.value
        elif str(series.dtype) in self.constants.DATETIME_INTERNAL_TYPES:
            series_result = PDTN.DATETIME.value
        elif str(series.dtype) in self.constants.OBJECT_INTERNAL_TYPES + self.constants.STRING_INTERNAL_TYPES:
            dftypes = series.apply(type)
            series_result = pd.Series([default_value] * len(series), dtype=np.uint8)
            series_result[dftypes == bool] = PDTN.BOOL.value
            series_result[dftypes == int] = PDTN.INT.value
            series_result[dftypes == float] = PDTN.FLOAT.value
            series_result[dftypes == list] = PDTN.LIST.value
            series_result[dftypes == dict] = PDTN.DICT.value
            series_result[dftypes == tuple] = PDTN.TUPLE.value
            series_result[dftypes == datetime.datetime] = PDTN.DATETIME.value
            series_result[series.isna()] = PDTN.NULL.value

            series_result[series == ""] = PDTN.MISSING.value
            mask_str = (series_result == default_value)
            series_result[mask_str] = PDTN.STRING.value

            new_mask = mask_str
            for i, (regex, type_value) in enumerate(
                    [
                        (regex_credit_card_compiled, PDTN.CREDIT_CARD_NUMBER.value),
                        (regex_phone_number_compiled, PDTN.PHONE_NUMBER.value),
                        (regex_address_compiled, PDTN.ADDRESS.value),

                        (regex_BAN_compiled, PDTN.BAN.value),
                        (regex_uuid_compiled, PDTN.UUID.value),
                        (regex_mac_address_compiled, PDTN.MAC_ADDRESS.value),
                        (regex_http_code_compiled, PDTN.HTTP_CODE.value),
                        (regex_gender_compiled, PDTN.GENDER.value),
                        (regex_social_security_number_compiled, PDTN.SOCIAL_SECURITY_NUMBER.value),
                        (regex_us_state_compiled, PDTN.US_STATE.value),
                        #
                        (regex_list_compiled, PDTN.LIST.value),
                        (regex_dict_compiled, PDTN.DICT.value),
                        (regex_tuple_compiled, PDTN.TUPLE.value),
                        #
                        (regex_zip_code_compiled, PDTN.ZIP_CODE.value),
                        (regex_email_compiled, PDTN.EMAIL.value),
                        (regex_ipv4_address_compiled, PDTN.IPV4_ADDRESS.value),
                        (regex_ipv6_address_compiled, PDTN.IPV6_ADDRESS.value),
                        (regex_full_url_compiled, PDTN.URL.value),
                        #
                        # # regex_date_compiled,
                        # # regex_time_compiled,

                        (regex_int_compiled, PDTN.INT.value),
                        (regex_decimal_compiled, PDTN.FLOAT.value),
                        (regex_boolean_compiled, PDTN.BOOLEAN.value),

                    ]):
                # This will only apply the regex over the data that did not match the previous regex
                next_mask = series[new_mask].str.match(regex)
                series_result.loc[next_mask | ~mask_str] = type_value
                new_mask = ~next_mask | ~mask_str
        else:
            series_result = PDTN.STRING.value

        return series_result

    def date_formats(self, series):
        return series.map(lambda v: hidateinfer.infer([v]))

    def metaphone(self, series):
        return self.to_string(series).map(jellyfish.metaphone, na_action='ignore')

    def double_metaphone(self, series):
        return self.to_string(series).map(doublemetaphone, na_action='ignore')

    def nysiis(self, series):
        return self.to_string(series).map(jellyfish.nysiis, na_action='ignore')

    def match_rating_codex(self, series):
        return self.to_string(series).map(jellyfish.match_rating_codex, na_action='ignore')

    def soundex(self, series):
        return self.to_string(series).map(jellyfish.soundex, na_action='ignore')

    def levenshtein(self, series, other):
        if isinstance(other, str):
            return self.to_string(series).map(lambda v: jellyfish.levenshtein_distance(v, other))
        else:
            col_name = series.name
            other_name = other.name
            dfd = self.to_string(series).to_frame()
            dfd[other_name] = self.to_string(other)

            return dfd.apply(lambda d: jellyfish.levenshtein_distance(d[col_name], d[other_name]), axis=1).rename(
                col_name)



import numpy as np
import pandas as pd
import sqlalchemy as sa
from dask.dataframe import from_delayed, from_pandas
from dask.delayed import delayed
from sqlalchemy import create_engine
from sqlalchemy import sql
from sqlalchemy.sql import elements

# Optimus plays defensive with the number of rows to be retrieved from the server so if a limit is not specified it will
# only will retrieve the LIMIT value
from optimus.engines.base.constants import NUM_PARTITIONS, LIMIT_TABLE
from optimus.engines.base.io.driver_context import DriverContext
from optimus.engines.base.io.factory import DriverFactory
from optimus.engines.base.io.properties import DriverProperties
from optimus.helpers.core import val_to_list
from optimus.helpers.logger import logger


class DaskBaseJDBC:
    """
    Helper for JDBC connections and queries
    """

    def __init__(self, host, database, user, password, port=None, driver=None, schema="public", oracle_tns=None,
                 oracle_service_name=None, oracle_sid=None, presto_catalog=None, cassandra_keyspace=None,
                 cassandra_table=None, bigquery_project=None, bigquery_dataset=None, op=None, sso=False):

        """
        Create the JDBC connection object
        :return:
        """

        self.op = op

        if host is None:
            host = "127.0.0.1"
            # RaiseIt.value_error(host, "host must be specified")

        if user is None:
            user = "root"
            # print("user",user)
            # RaiseIt.value_error(user, "user must be specified")

        if database is None:
            database = ""

        self.db_driver = driver
        self.oracle_sid = oracle_sid
        self.cassandra_keyspace = cassandra_keyspace
        self.cassandra_table = cassandra_table

        self.driver_context = DriverContext(DriverFactory.get(self.db_driver))
        self.driver_properties = self.driver_context.properties()

        if port is None:
            self.port = self.driver_properties.value["port"]
        else:
            self.port = port

        self.driver_option = self.driver_properties.value["java_class"]

        self.uri = self.driver_context.uri(
            user=user,
            password=password,
            driver=driver,
            host=host,
            port=str(self.port),
            database=database,
            schema=schema,
            oracle_tns=oracle_tns,
            oracle_sid=oracle_sid,
            oracle_service_name=oracle_service_name,
            presto_catalog=presto_catalog,
            bigquery_project=bigquery_project,
            bigquery_dataset=bigquery_dataset,
            sso=sso
        )

        self.database = database
        self.user = user
        self.password = password
        self.schema = schema
        print(self.uri)
        logger.print(self.uri)

    def tables(self, schema=None, database=None, limit=None):
        """
        Return all the tables in a database
        :return:
        """

        # Override the schema used in the constructor
        # if database is None:
        #     database = self.database
        #
        # if schema is None:
        #     schema = self.schema
        # query = self.driver_context.table_names_query(schema=schema, database=database)
        # df = self.execute(query, limit)
        # return df.display(limit)

        engine = create_engine(self.uri)
        return engine.table_names()

    @property
    def table(self):
        """
        Print n rows of every table in a database
        :return: Table Object
        """
        return Table(self)

    def table_to_df(self, table_name: str, columns="*", partition_column =None,limit=None, n_partitions=1):
        """
        Return cols as Spark data frames from a specific table
        :type table_name: object
        :param columns:
        :param limit: how many rows will be retrieved
        """
        db_table = table_name

        if limit == "all":
            query = self.driver_context.count_query(db_table=db_table)
            count = self.execute(query, "all").first()[0]

            # We want to count the number of rows to warn the users how much it can take to bring the whole data
            print(str(int(count)) + " rows")

        if columns == "*":
            columns_sql = "*"
        else:
            columns = val_to_list(columns)
            columns_sql = ",".join(columns)

        query = "SELECT " + columns_sql + " FROM " + db_table

        logger.print(query)

        dfd = self.execute(query, limit, n_partitions=n_partitions)
        # Bring the data to local machine if not every time we call an action is going to be
        # retrieved from the remote server
        # dfd = dfd.run()
        # dfd = dask_pandas_to_dask_cudf(dfd)
        # print(dfd)
        # print(self.op.F.dask_to_compatible(dfd).head())
        return self.op.create.dataframe(self.op.F.dask_to_compatible(dfd))

    def execute(self, query, limit=None, n_partitions: int = NUM_PARTITIONS, partition_column: str = None,
                table_name=None):
        """
        Execute a SQL query
        :param limit: default limit the whole query. We play defensive here in case the result is a big chunk of data
        :param n_partitions:
        :param partition_column:
        :param query: SQL query string
        :param table_name:
        :return:
        """

        dfd = DaskBaseJDBC.read_sql_table(table_name=table_name, uri=self.uri, index_col=partition_column,
                                          npartitions=n_partitions, query=query)

        return self.op.create.dataframe(self.op.F.dask_to_compatible(dfd))

    @staticmethod
    def read_sql_table(
            table_name,
            uri,
            index_col=None,
            divisions=None,
            npartitions=None,
            limits=None,
            columns=None,
            bytes_per_chunk=256 * 2 ** 20,
            head_rows=5,
            schema=None,
            meta=None,
            engine_kwargs=None,
            query=None,
            **kwargs
    ):

        engine_kwargs = {} if engine_kwargs is None else engine_kwargs
        engine = sa.create_engine(uri, **engine_kwargs)
        m = sa.MetaData()

        # print("table", table)
        if isinstance(table_name, str):
            table_name = sa.Table(table_name, m, autoload=True, autoload_with=engine, schema=schema)

            columns = (
                [(table_name.columns[c] if isinstance(c, str) else c) for c in columns]
                if columns
                else list(table_name.columns)
            )
        index = None
        if index_col:
            # raise ValueError("Must specify index column to partition on")
            # else:

            index = table_name.columns[index_col] if isinstance(index_col, str) else index_col
            if not isinstance(index_col, (str, elements.Label)):
                raise ValueError(
                    "Use label when passing an SQLAlchemy instance as the index (%s)" % index
                )
            if divisions and npartitions:
                raise TypeError("Must supply either divisions or npartitions, not both")

            if index_col not in columns:
                columns.append(
                    table_name.columns[index_col] if isinstance(index_col, str) else index_col
                )

            if isinstance(index_col, str):
                kwargs["index_col"] = index_col
            else:
                # function names get pandas auto-named
                kwargs["index_col"] = index_col.name
        parts = []

        if meta is None:
            # derive metadata from first few rows
            # q = sql.select(columns).limit(head_rows).select_from(table)
            head = pd.read_sql(query, engine, **kwargs)

            if head.empty:
                # no results at all
                # name = table_name.name
                # schema = table_name.schema
                # head = pd.read_sql_table(name, uri, schema=schema, index_col=index_col)

                return from_pandas(head, npartitions=1)

            bytes_per_row = (head.memory_usage(deep=True, index=True)).sum() / head_rows
            meta = head.iloc[:0]
            # print(list(head.columns.values))
        else:
            if divisions is None and npartitions is None:
                raise ValueError(
                    "Must provide divisions or npartitions when using explicit meta."
                )

        # if divisions is None and index_col is not None:
        if divisions is None:
            # print(index)
            # print("LIMITS",limits)
            if index is not None and limits is None:
                # calculate max and min for given index
                q = sql.select([sql.func.max(index), sql.func.min(index)]).select_from(
                    table_name
                )
                minmax = pd.read_sql(q, engine)
                maxi, mini = minmax.iloc[0]
                dtype = minmax.dtypes["max_1"]
            elif index is None and npartitions:
                # User for Limit offset
                mini = 0
                q = f"SELECT COUNT(*) AS count FROM ({query}) AS query"
                maxi = pd.read_sql(q, engine)["count"][0]
                limit = maxi / npartitions
                dtype = pd.Series((mini, maxi,)).dtype

                #  Use for ntile calculation
                # mini = 0
                # maxi = npartitions
                # print(pd.Series((mini, maxi,)))
                # dtype = pd.Series((mini, maxi,)).dtype
                # ntile_columns = ", ".join(list(head.columns.values))

            else:
                mini, maxi = limits
                dtype = pd.Series(limits).dtype

            if npartitions is None:
                q = sql.select([sql.func.count(index)]).select_from(table_name)
                count = pd.read_sql(q, engine)["count_1"][0]
                npartitions = int(round(count * bytes_per_row / bytes_per_chunk)) or 1

            if dtype.kind == "M":
                divisions = pd.date_range(
                    start=mini,
                    end=maxi,
                    freq="%iS" % ((maxi - mini).total_seconds() / npartitions),
                ).tolist()
                divisions[0] = mini
                divisions[-1] = maxi
            elif dtype.kind in ["i", "u", "f"]:
                divisions = np.linspace(mini, maxi, npartitions + 1).tolist()
            # else:
            #     print(dtype)
            #     raise TypeError(
            #         'Provided index column is of type "{}".  If divisions is not provided the '
            #         "index column type must be numeric or datetime.".format(dtype)
            #     )

            lowers, uppers = divisions[:-1], divisions[1:]
            for i, (lower, upper) in enumerate(zip(lowers, uppers)):

                if index_col:
                    if i == len(lowers) - 1:
                        where = f" WHERE {index} > {lower} AND {index} <= {upper}"
                    else:
                        where = f" WHERE {index} >= {lower} AND {index} < {upper}"
                    q = query + where
                else:
                    if i == len(lowers) - 1:
                        where = f" LIMIT {int(upper) - int(lower)} OFFSET {int(lower)}"
                    else:
                        where = f" LIMIT {int(limit)} OFFSET {int(lower)}"
                    q = query + where

                    # Ntile calculation
                    # print("Using Ntile query")
                    # ntile_column = "temp"
                    # ntile_sql = f"SELECT *, NTILE({npartitions}) OVER(ORDER BY id DESC) AS {ntile_column} FROM ({query}) AS t";
                    # q = f"SELECT {ntile_columns} FROM ({ntile_sql}) AS r"
                    # if i == len(lowers) - 1:
                    #     where = f" WHERE {ntile_column} > {lower} AND {ntile_column} <= {upper}"
                    # else:
                    #     where = f" WHERE {ntile_column} >= {lower} AND {ntile_column} < {upper}"
                    # q = q + where

                    # table = "test_data"
                    # q = f'SELECT {ntile_columns} FROM {table} WHERE  NTILE({npartitions}) OVER (ORDER BY {ntile_column}) = i'

                # When we do not have and index
                parts.append(
                    delayed(DaskBaseJDBC._read_sql_chunk)(
                        q, uri, meta, engine_kwargs=engine_kwargs, **kwargs
                    )
                )
        else:

            # JDBC._read_sql_chunk(q, uri, meta, engine_kwargs=engine_kwargs, **kwargs)

            parts.append(
                delayed(DaskBaseJDBC._read_sql_chunk)(
                    query, uri, meta, engine_kwargs=engine_kwargs, **kwargs
                )
            )

        engine.dispose()

        return from_delayed(parts, meta, divisions=divisions)

    @staticmethod
    def _read_sql_chunk(q, uri, meta, engine_kwargs=None, **kwargs):
        import sqlalchemy as sa

        engine_kwargs = engine_kwargs or {}
        engine = sa.create_engine(uri, **engine_kwargs)
        df = pd.read_sql(q, engine, **kwargs)
        engine.dispose()

        if df.empty:
            return meta
        else:
            # df.reset_index()
            return df.astype(meta.dtypes.to_dict(), copy=False)

    def df_to_table(self, df, table, mode="overwrite"):
        """
        Sends a dataframe to the database
        :param df:
        :param table:
        :param mode
        :return:
        """
        # Parse array and vector to string. JDBC can not handle this data types
        columns = df.cols.names("*", by_dtypes=["array", "vector"])
        df = df.cols.cast(columns, "str")

        conf = df.write \
            .format(
            "jdbc" if not self.db_driver == DriverProperties.CASSANDRA.value["name"] else
            DriverProperties.CASSANDRA.value["java_class"]) \
            .mode(mode) \
            .option("url", self.uri) \
            .option("dbtable", table) \
            .option("user", self.user) \
            .option("password", self.password)

        if self.db_driver == DriverProperties.ORACLE.value["name"]:
            conf.option("driver", self.driver_option)
        conf.save()

    @staticmethod
    def _limit(df, limit=None):
        """
        Handle limit defensive so we do not retrieve the whole at we explicit want
        :param limit:
        :param df:
        :return a limited DataFrame if specified
        """
        # we use a default limit here in case the query will return a huge chunk of data
        if limit is None:
            return df.limit(LIMIT_TABLE)
        elif limit == "all":
            return df
        else:
            return df.limit(int(limit))


class Table:
    def __init__(self, db):
        self.db = db

    def show(self, table_names="*", limit=None):
        db = self.db

        if table_names == "*":
            table_names = db.tables()
        else:
            table_names = val_to_list(table_names)

        print("Total Tables:" + str(len(table_names)))

        for table_name in table_names:
            db.table_to_df(table_name, "*", limit) \
                .table(title=table_name)



import collections
import functools
import glob
import ntpath
import os
import random
import re
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Union
from urllib.parse import unquote
from urllib.request import Request, urlopen

import humanize
import numpy as np
import six
from fastnumbers import isint, isfloat

from optimus import ROOT_DIR
from optimus.helpers.constants import DATE_FORMAT_ITEMS, DATE_FORMAT_ITEMS_MONTH, PYTHON_DATE_TO_FORMAT
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.logger import logger
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_dict, is_list_with_dicts, is_numeric, is_url


def _list_variables(ins, namespace=None):
    if not namespace:
        try:
            import IPython
            namespace = IPython.get_ipython().user_global_ns
        except:
            pass

    return [obj for obj in (namespace or []) if isinstance(namespace[obj], ins) and not obj.startswith("_")]


def list_engines(namespace=None):
    from optimus.engines.base.engine import BaseEngine
    return _list_variables(BaseEngine, namespace)


def list_dataframes(namespace=None):
    from optimus.engines.base.basedataframe import BaseDataFrame
    return _list_variables(BaseDataFrame, namespace)


def list_clusters(namespace=None):
    from optimus.engines.base.stringclustering import Clusters
    return _list_variables(Clusters, namespace)


def list_connections(namespace=None):
    from optimus.engines.base.dask.io.jdbc import DaskBaseJDBC
    from optimus.engines.base.io.connect import Connection
    return _list_variables((Connection, DaskBaseJDBC), namespace)


def list_file_connections(namespace=None):
    from optimus.engines.base.io.connect import Connection
    return _list_variables(Connection, namespace)


def list_database_connections(namespace=None):
    from optimus.engines.base.dask.io.jdbc import DaskBaseJDBC
    return _list_variables(DaskBaseJDBC, namespace)


def random_int(n=5):
    """
    Create a random string of ints
    :return:
    """
    return str(random.randint(1, 10 ** n))


def collect_as_list(df):
    return df.rdd.flatMap(lambda x: x).collect()


def filter_list(val, index=0):
    """
    Convert a list to None, int, str or a list filtering a specific index
    [] to None
    ['test'] to test

    :param val:
    :param index:
    :return:
    """
    if len(val) == 0:
        return None
    else:
        return one_list_to_val([column[index] for column in val])


def absolute_path(files, format="posix"):
    """
    User project base folder to construct and absolute path
    :param files: path files
    :param format: posix or uri
    :return:
    """
    files = val_to_list(files)
    result = None
    if format == "uri":
        result = [Path(ROOT_DIR + file).as_uri() for file in files]
    elif format == "posix":
        result = [Path(ROOT_DIR + file).as_posix() for file in files]
    else:
        RaiseIt.value_error(format, ["posix", "uri"])

    result = one_list_to_val(result)
    return result


def format_path(path, format="posix"):
    """
    Format a path depending fo the operative system
    :param path:
    :param format:
    :return:
    """
    if format == "uri":
        result = Path(path).as_uri()
    elif format == "posix":
        result = Path(path).as_posix()
    return result


def java_version():
    version = subprocess.check_output(['java', '-version'], stderr=subprocess.STDOUT)
    pattern = r'\"(\d+\.\d+).*\"'
    print(re.search(pattern, version).groups()[0])


def setup_google_colab():
    """
    Check if we are in Google Colab and setup it up
    :return:
    """
    from optimus.helpers.constants import JAVA_PATH_COLAB
    from optimus.engines.spark.constants import SPARK_PATH_COLAB
    from optimus.engines.spark.constants import SPARK_URL
    from optimus.engines.spark.constants import SPARK_FILE

    IN_COLAB = 'google.colab' in sys.modules

    if IN_COLAB:
        if not os.path.isdir(JAVA_PATH_COLAB) or not os.path.isdir(SPARK_PATH_COLAB):
            print("Installing Optimus, Java8 and Spark. It could take 3 min...")
            commands = [
                "apt-get install openjdk-8-jdk-headless -qq > /dev/null",
                "wget -q {SPARK_URL}".format(SPARK_URL=SPARK_URL),
                "tar xf {SPARK_FILE}".format(SPARK_FILE=SPARK_FILE)
            ]

            cmd = " && ".join(commands)

            p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
            p_stdout = p.stdout.read().decode("ascii")
            p_stderr = p.stderr.read().decode("ascii")
            print(p_stdout, p_stderr)

        else:
            print("Settings env vars")
            # Always configure the env vars

            os.environ["JAVA_HOME"] = JAVA_PATH_COLAB
            os.environ["SPARK_HOME"] = SPARK_PATH_COLAB


def is_pyarrow_installed():
    """
    Check if pyarrow is installed
    :return:
    """
    try:
        import pyarrow
        have_arrow = True
    except ImportError as e:
        print(e)
        have_arrow = False
    return have_arrow


def check_env_vars(env_vars):
    """
    Check if a environment var exist
    :param env_vars: Environment var name
    :return:
    """

    for env_var in env_vars:
        if env_var in os.environ:
            logger.print(env_var + "=" + os.environ.get(env_var))
        else:
            logger.print(env_var + " is not set")


# Reference https://nvie.com/posts/modifying-deeply-nested-structures/

def transform_date_format(format: str):
    """
    Transform a date format like `yyyy/mm/dd` to a compatible one like `%Y/%m/%d`
    """

    has_time = "H" in format or "h" in format

    items = DATE_FORMAT_ITEMS if has_time else DATE_FORMAT_ITEMS_MONTH

    for i, j in items:
        reg = re.compile(f"(?<![%])(?<!%-){i}")
        format = re.sub(reg, j, format)

    reg = re.compile("(?<![%])(?<!%-)[A-z]")
    return re.sub(reg, lambda a: f"%{a.group()}", format)


def ellipsis(data, length=20):
    """
    Add a "..." if a string y greater than a specific length
    :param data:
    :param length: length taking into account to cut the string
    :return:
    """
    data = str(data)
    return (data[:length] + '..') if len(data) > length else data


def create_buckets(lower_bound, upper_bound, bins):
    """
    Create a dictionary with bins
    :param lower_bound: low range
    :param upper_bound: high range
    :param bins: number of buckets
    :return:
    """
    range_value = (upper_bound - lower_bound) / bins
    low = lower_bound

    buckets = []

    if bins == 1:
        buckets.append({"lower": low, "upper": low + 1, "bucket": 0})
    else:
        for i in range(0, bins):
            high = low + range_value
            buckets.append({"lower": low, "upper": high, "bucket": i})
            low = high

        # Ensure that the upper bound is exactly the higher value.
        # Because floating point calculation it can miss the upper bound in the final sum

        buckets[bins - 1]["upper"] = upper_bound
    return buckets


def deep_sort(obj):
    """
    Recursively sort list or dict nested lists
    """

    if isinstance(obj, dict):
        _sorted = {}
        for key in sorted(obj):
            _sorted[key] = deep_sort(obj[key])

    elif isinstance(obj, list):
        new_list = []
        for val in obj:
            new_list.append(deep_sort(val))
        _sorted = sorted(new_list)

    else:
        _sorted = obj

    return _sorted


def update_dict(d, u):
    """
    Update only the given keys
    :param d:
    :param u:
    :return:
    """
    # python 3.8+ compatibility
    try:
        collectionsAbc = collections.abc
    except ModuleNotFoundError:
        collectionsAbc = collections

    for k, v in six.iteritems(u):
        dv = d.get(k, {})
        if not isinstance(dv, collectionsAbc.Mapping):
            d[k] = v
        elif isinstance(v, collectionsAbc.Mapping):
            d[k] = update_dict(dv, v)
        else:
            d[k] = v
    return d


def reduce_mem_usage(df, categorical=True, categorical_threshold=50, verbose=False):
    """
    Change the columns datatype to reduce the memory usage.
    :param df: Dataframe to reduce the memory usage
    :param categorical:
    :param categorical_threshold: Max number to categorical features to me considered a categorical column
    :param verbose:
    :return:
    """

    # Reference https://www.kaggle.com/arjanso/reducing-dataframe-memory-size-by-65/notebook

    start_mem_usg = df.size()
    dfd = df.data
    ints = df.cols.exec_agg(dfd.applymap(isint).sum(), True)
    floats = df.cols.exec_agg(dfd.applymap(isfloat).sum(), True)
    nulls = df.cols.exec_agg(dfd.isnull().sum(), True)
    total_rows = len(dfd)

    columns_dtype = {}
    for x, y in ints.items():

        if ints[x] == nulls[x]:
            dtype = "object"
        elif floats[x] == total_rows:
            dtype = "numerical"
        elif total_rows <= ints[x] + nulls[x]:
            dtype = "numerical"
        else:
            dtype = "object"
        columns_dtype[x] = dtype

    numerical_int = [col for col, dtype in columns_dtype.items() if dtype == "numerical"]
    final = {}

    if len(numerical_int) > 0:
        min_max = df.cols.range(numerical_int)

        import numpy as np
        for col_name in min_max.keys():
            _min = min_max[col_name]["min"]
            _max = min_max[col_name]["max"]
            if _min >= 0:
                if _max < 255:
                    final[col_name] = np.uint8
                elif _max < 65535:
                    final[col_name] = np.uint16
                elif _max < 4294967295:
                    final[col_name] = np.uint32
                else:
                    final[col_name] = np.uint64
            else:
                if _min > np.iinfo(np.int8).min and _max < np.iinfo(np.int8).max:
                    final[col_name] = np.int8
                elif _min > np.iinfo(np.int16).min and _max < np.iinfo(np.int16).max:
                    final[col_name] = np.int16
                elif _min > np.iinfo(np.int32).min and _max < np.iinfo(np.int32).max:
                    final[col_name] = np.int32
                elif _min > np.iinfo(np.int64).min and _max < np.iinfo(np.int64).max:
                    final[col_name] = np.int64
            # print(final[col_name])

    object_int = [col for col, dtype in columns_dtype.items() if dtype == "object"]
    if len(object_int) > 0:
        count_values = df.cols.value_counts(object_int)

    # if categorical is True:
    #     for col_name in object_int:
    #         if len(count_values[col_name]) <= categorical_threshold:
    #             final[col_name] = "category"

    dfd = dfd.astype(final)
    mem_usg = dfd.size

    if verbose is True:
        print("Memory usage after optimization:", humanize.naturalsize(start_mem_usg))
        print("Memory usage before optimization is: ", humanize.naturalsize(mem_usg))
        print(round(100 * mem_usg / start_mem_usg), "% of the initial size")

    return df.new(dfd, meta=df.meta)


def downloader(url, file_format):
    """
    Send the request to download a file
    """

    def write_file(response, file, chunk_size=8192):
        """
        Load the data from the http request and save it to disk
        :param response: data returned from the server
        :param file:
        :param chunk_size: size chunk size of the data
        :return:
        """
        total_size = response.headers['Content-Length'].strip() if 'Content-Length' in response.headers else 100
        total_size = int(total_size)
        bytes_so_far = 0

        while 1:
            chunk = response.read(chunk_size)
            bytes_so_far += len(chunk)
            if not chunk:
                break
            file.write(chunk)
            total_size = bytes_so_far if bytes_so_far > total_size else total_size

        return bytes_so_far

    # try to infer the file format using the file extension
    if file_format is None:
        filename, file_format = os.path.splitext(url)
        file_format = file_format.replace('.', '')

    i = url.rfind('/')
    data_name = url[(i + 1):]

    headers = {"User-Agent": "Optimus Data Downloader/1.0"}

    req = Request(url, None, headers)

    logger.print("Downloading %s from %s", data_name, url)

    # It seems that avro need a .avro extension file
    with tempfile.NamedTemporaryFile(suffix="." + file_format, delete=False) as f:
        bytes_downloaded = write_file(urlopen(req), f)
        path = f.name

    if bytes_downloaded > 0:
        logger.print("Downloaded %s bytes", bytes_downloaded)

    logger.print("Creating DataFrame for %s. Please wait...", data_name)

    return path


def unquote_path(path):
    return unquote(path)


@functools.lru_cache(maxsize=128)
def prepare_path(path, file_format=None):
    """d
    Helper to return the file to be loaded and the file name.
    This will memoise
    :param path: Path to the file to be loaded
    :param file_format: format file
    :return:
    """
    r = []
    if is_url(path):
        file = downloader(path, file_format)
        file_name = ntpath.basename(path)
        r = [(file, file_name,)]
    else:
        for file_name in glob.glob(path, recursive=True):
            r.append((file_name, ntpath.basename(file_name),))
    if len(r) == 0:
        raise Exception("File not found")
    return r


def prepare_path_local(path):
    """
    Helper create the folder of the file to be saved
    :param path: Path to the file to be saved
    """
    path = os.path.join(Path().absolute(), path)
    os.makedirs(os.path.dirname(path), exist_ok=True)


def path_is_local(path):
    """
    Check if path is local
    :param path: Path to the file to be saved
    """
    from optimus.helpers.constants import Schemas
    if path.startswith((*Schemas.list(),)):
        return path.startswith(Schemas.FILE.value)
    return True


def prepare_url_schema(path: str, schema="https", force=False):
    """
    Prepare a url and include a schema if necessary.
    :param path: Url path to prepare.
    :param schema: schema to include if none is found in 'path'.
    :param force: forces the schema passed to 'schema'.
    """

    if not schema.endswith("://"):
        schema += "://"

    import hiurlparser
    parsed_url = hiurlparser.parse_url(path)
    found_schema = parsed_url["protocol"] if parsed_url else None

    if found_schema is not None and not force:
        return path
    elif found_schema is None:
        return schema + path
    else:
        return schema + path[len(found_schema + "://"):]


def month_names(directive="%B"):
    """
    Gets an array with month names
    """
    import datetime
    return [datetime.datetime.strptime(str(n + 1), '%m').strftime(directive) for n in range(12)]


def weekday_names(directive="%B"):
    """
    Gets an array with month names
    """
    import datetime
    return [datetime.datetime.strptime(str(n + 1), '%m').strftime(directive) for n in range(7)]


def match_date(value, error=False):
    """
    Create a regex from a string with a date format like
    `dd/MM/yyyy hh:mm:ss-sss mi` or `%d/%m/%Y`.
    :param value: Value to test
    :return: String regex
    """
    formats = ["d", "dd", "M", "MM", "MMM", "MMMM", "MMMMM", "yy", "yyyy", "w",
               "ww", "www", "wwww", "wwwww", "h", "hh", "H", "HH", "kk", "k",
               "m", "mm", "s", "ss", "sss", "xxx", "/", ":", "-", " ", "+", "|", "a",
               "mi"]
    formats.sort(key=len, reverse=True)

    if "%" in value:
        for i, j in PYTHON_DATE_TO_FORMAT.items():
            value = value.replace(i, j)

    result = []

    start = 0

    end = len(value)

    while start < end:
        found = False
        for f in formats:
            ignore_case = ("h" not in f.lower() and "m" not in f.lower())
            _value = value.lower() if ignore_case else value
            if _value.startswith(f, start):
                start = start + len(f)
                result.append(f)
                found = True
                break
        if found is False:
            error_str = f"{value[start]} is not a valid date format. ({value})"
            if error:
                raise ValueError(error_str)
            else:
                logger.warn(error_str)
                return False

    exprs = []
    for f in result:
        # Separators
        if f in ["/", ":", "-", " ", "|", "+", " "]:
            exprs.append("\\" + f)

        # Day
        # d  -> 1 ... 31
        # dd -> 01 ... 31

        elif f == "d":
            exprs.append("((3[0-1])|([1-2][0-9])|(0?[1-9]))")
        elif f == "dd":
            exprs.append("((3[0-1])|([1-2][0-9])|(0[1-9]))")

        # Month
        # M    -> 1 ... 12
        # MM   -> 01 ... 12
        # MMM  -> Jan, Feb, Sep
        # MMMM -> January, February, September
        elif f == "M":
            exprs.append("((1[0-2])|(0?[1-9]))")
        elif f == "MM":
            exprs.append("((1[0-2])|(0[1-9]))")
        elif f == "MMM":
            exprs.append(f"({'|'.join(month_names('%b'))})")
        elif f in ["MMMM", "MMMMM"]:
            exprs.append(f"({'|'.join(month_names('%B'))})")

        # Weekday
        # w    -> 0 ... 6
        # ww   -> 00 ... 06
        # www  -> Mon, Tues, Sat
        # wwww -> Monday, Tuesday, Saturday
        elif f == "w":
            exprs.append("(0?[0-6])")
        elif f == "ww":
            exprs.append("(0[0-6])")
        elif f == "www":
            exprs.append(f"({'|'.join(weekday_names('%a'))})")
        elif f in ["wwww", "wwwww"]:
            exprs.append(f"({'|'.join(weekday_names('%A'))})")

        # Year
        # yy   -> 00 ... 99
        # yyyy -> 0000 ... 9999
        elif f == "yy":
            exprs.append("([0-9]{2})")
        elif f == "yyyy":
            exprs.append("([0-9]{4})")

            # Hours
        # h  -> 1,2 ... 12
        # hh -> 01,02 ... 12
        # H  -> 0,1 ... 23
        # HH -> 00,01 ... 23
        # k  -> 1,2 ... 24
        # kk -> 01,02 ... 24
        elif f == "h":
            exprs.append("((1[0-2])|(0?[1-9]))")
        elif f == "hh":
            exprs.append("((1[0-2])|(0[1-9]))")
        elif f == "H":
            exprs.append("((0?[0-9])|(1[0-9])|(2[0-3])|[0-9])")
        elif f == "HH":
            exprs.append("((0[0-9])|(1[0-9])|(2[0-3])|[0-9])")
        elif f == "k":
            exprs.append("((0?[1-9])|(1[0-9])|(2[0-4])|[1-9])")
        elif f == "kk":
            exprs.append("((0[1-9])|(1[0-9])|(2[0-4]))")

        # Minutes
        # m  -> 0 ... 59
        # mm -> 00 .. 59
        elif f == "m":
            exprs.append("[1-5]?[0-9]")
        elif f == "mm":
            exprs.append("[0-5][0-9]")

        # Seconds
        # s  -> 0 ... 59
        # ss -> 00 .. 59
        elif f == "s":
            exprs.append("[1-5]?[0-9]")
        elif f == "ss":
            exprs.append("[0-5][0-9]")

        # Milliseconds
        # sss -> 0 ... 999
        elif f == "sss":
            exprs.append("([0-9]{3})")

        # Extras
        # mi or a -> Meridian indicator (AM am Am) (PM pm Pm) (m M)
        elif f in ["mi", "a"]:
            exprs.append(r"([A|P|a|p]?.?\s?)[M|m].?")

    return r"".join(exprs)


def df_dicts_equal(df1, df2, decimal: Union[int, bool] = True, assertion=False):
    if decimal is True:
        decimal = 7
    for k in df1:
        try:
            atol = pow(10.0, -decimal) if is_numeric(decimal) else 0
            np.testing.assert_allclose(df1[k], df2[k], atol=atol, equal_nan=True)
        except AssertionError as e:
            if df1[k] == df2[k]:
                return True
            if assertion:
                logger.warn(f"AssertionError on column {k}")
                raise e
            return False
        except Exception:
            if df1[k] != df2[k]:
                if assertion:
                    raise AssertionError(f"Dataframes are not equal on column '{k}'")
                return False
    return True


def results_equal(r1, r2, decimal: Union[int, bool] = True, assertion=False):
    if decimal is True:
        decimal = 7

    try:

        if hasattr(r1, "__len__") and len(r1) != len(r2):
            raise AssertionError(f"Lengths '{len(r1)}' and '{len(r2)}' do not match ('{r1}', '{r2}')")

        matching = None

        if is_dict(r1):

            if not is_dict(r2):
                raise AssertionError(f"Types '{str(type(r1))}' and '{str(type(r2))}' do not match")

            matching = True

            for key in r1:
                if not results_equal(r1[key], r2.get(key, None), decimal, assertion):
                    matching = False
                    break

        if is_list_with_dicts(r1):
            for e1, e2 in zip(r1, r2):
                if not results_equal(e1, e2, decimal, assertion):
                    matching = False
                    break

        if matching is not None:
            return matching

        try:
            np.testing.assert_almost_equal(r1, r2, decimal=decimal)
        except AssertionError as e:
            raise e
        except Exception:
            if r1 != r2:
                raise AssertionError(f"'{r1}' and '{r2}' do not match")
            else:
                return True

    except Exception as e:
        if assertion:
            raise e
        return False

    return True


# Taken from https://github.com/Kemaweyan/singleton_decorator/
class _SingletonWrapper:
    """
    A singleton wrapper class. Its instances would be created
    for each decorated class.
    """

    def __init__(self, cls):
        self.__wrapped__ = cls
        self._instance = None

    def __call__(self, *args, **kwargs):
        """Returns a single instance of decorated class"""
        if self._instance is None:
            self._instance = self.__wrapped__(*args, **kwargs)
        return self._instance


def singleton(cls):
    """
    A singleton decorator. Returns a wrapper objects. A call on that object
    returns a single instance object of decorated class. Use the __wrapped__
    attribute to access decorated class directly in unit tests
    """
    return _SingletonWrapper(cls)



from optimus.tests.base import TestBase


class TestUDFPandas(TestBase):
    load = {"path": "examples/data/DCIGNP2AYL.txt", "n_rows": 20}

    def test_udf(self):
        from optimus.functions import F

        df = self.df

        def haversine(lat1, lon1, lat2, lon2):    
            MILES = 3959
            lat1, lon1, lat2, lon2 = map(F.radians, [lat1, lon1, lat2, lon2])
            dlat = lat2 - lat1 
            dlon = lon2 - lon1
            a = F.sin(dlat / 2)**2 + F.cos(lat1) * F.cos(lat2) * F.sin(dlon / 2)**2
            c = 2 * F.asin(F.sqrt(a))  
            total_miles = MILES * c
            return total_miles

        df["distance"] = haversine(40.671, -73.985, df['latitude'], df['longitude'])

        self.assertIn("distance", df.cols.names())

        expected_dict = {"distance": [139.6071899, 139.7468976, 142.1910502, 137.275546, 139.6845834, 142.511427,
                                      139.2236552, 137.5044141, 141.9774911, 138.9195547, 141.1156307, 137.320386,
                                      136.9785681, 139.8973388, 142.0501961, 141.0701359, 141.6853331, 141.7333734,
                                      139.0531985, 139.5583856]}

        self.assertTrue(df["distance"].equals(expected_dict, decimal=True, assertion=True))


class TestUDFDask(TestUDFPandas):
    config = {'engine': 'dask', 'n_partitions': 1}


class TestUDFPartitionDask(TestUDFPandas):
    config = {'engine': 'dask', 'n_partitions': 2}


try:
    import cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestUDFCUDF(TestUDFPandas):
        config = {'engine': 'cudf'}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestUDFDC(TestUDFPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 1}


try:
    import dask_cudf # pyright: reportMissingImports=false
except:
    pass
else:
    class TestUDFPartitionDC(TestUDFPandas):
        config = {'engine': 'dask_cudf', 'n_partitions': 2}

try:
    import pyspark
except:
    pass
else:
    class TestUDFSpark(TestUDFPandas):
        config = {'engine': 'spark'}

try:
    import vaex
except:
    pass
else:
    class TestUDFVaex(TestUDFPandas):
        config = {'engine': 'vaex'}


import time
from abc import ABC

from glom import assign

from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns
from optimus.helpers.core import one_list_to_val
from optimus.helpers.functions import update_dict
from optimus.helpers.json import dump_json
from optimus.helpers.types import *
from optimus.infer import is_list
from optimus.profiler.constants import MAX_BUCKETS


class BaseProfile(ABC):
    """Base class for all profile implementations"""

    def __init__(self, root: 'DataFrameType'):
        self.root = root

    def _columns_summary(self):
        df = self.root

        data_types_list = list(set(df.cols.data_type("*", tidy=False)["data_type"].values()))

        data_types = df.cols.data_type("*", tidy=False)["data_type"]

        return {
            'cols_count': df.cols.count(),
            'data_types_list': data_types_list,
            'total_count_data_types': len(set([i for i in data_types.values()]))
        }

    def summary(self, cols="*"):

        df = self.root

        return Meta.get(df.meta, "profile.summary")

    def columns(self, cols="*"):

        df = self.root
        cols = parse_columns(df, cols) if cols else []

        if is_list(cols):
            cols = [Meta.get(df.meta, f"profile.columns.{col}") for col in cols]
        else:
            cols = Meta.get(df.meta, f"profile.columns.{cols}")

        return one_list_to_val(cols)

    def data_type(self, cols="*"):

        df = self.root
        cols = parse_columns(df, cols) if cols else []

        dtype = [Meta.get(
            df.meta, f"profile.columns.{col}.stats.inferred_data_type.data_type") for col in cols]

        return one_list_to_val(dtype)

    def __call__(self, cols="*", bins: int = MAX_BUCKETS, force_hist=None, output: str = None, flush: bool = False,
                 size=False) -> dict:
        """
        Returns a dict the profile of the dataset
        :param cols: Columns to get the profile from
        :param bins: Number of buckets
        :param force_hist: Force histogram on selected columns
        :param output: Output format
        :param flush: Flushes the cache of the whole profile to process it again
        :param size: get the dataframe size in memory. Use with caution this could be slow for big data frames.
        :return:
        """

        df = self.root

        meta = self.root.meta
        profile = Meta.get(meta, "profile")

        if not profile:
            flush = True

        if cols or flush:
            cols = parse_columns(df, cols) if cols else []
            transformations = Meta.get(meta, "transformations") or []

            calculate = False

            if flush or len(transformations):
                calculate = True

            else:
                for col in cols:
                    if col not in profile["columns"]:
                        calculate = True

            if calculate:
                df = df.profile._calculate(cols, bins, force_hist, flush, size)
                profile = Meta.get(df.meta, "profile")
                self.root.meta = df.meta
            profile["columns"] = {key: profile["columns"][key] for key in cols}
        
        profile["summary"].update(df.profile._columns_summary())
        
        self.root.meta = Meta.set(df.meta, "profile.summary", profile["summary"])

        if output == "json":
            profile = dump_json(profile)

        return profile

    def _calculate(self, cols="*", bins: int = MAX_BUCKETS, force_hist=None, flush: bool = False, size=False):
        """
        Returns a new dataframe with the profile data added to the meta property
        :param cols: "*", column name or list of column names to be processed.
        :param bins:
        :param flush:
        :param size: get the dataframe size in memory. Use with caution this could be slow for big data frames.
        :return:
        """

        _t = time.process_time()
        profiler_time = {"hist": {}, "frequency": {}, "count_mismatch": {}}

        df = self.root
        meta = df.meta

        force_hist = df.cols.names(force_hist) if force_hist is not None else []

        if flush is False:
            cols_to_profile = df._cols_to_profile(cols) or []
        else:
            cols_to_profile = parse_columns(df, cols) or []

        profiler_data = Meta.get(meta, "profile")

        is_cached = profiler_data is not None

        if profiler_data is None:
            profiler_data = {}
        cols_data_types = None

        profiler_time["beginning"] = {"elapsed_time": time.process_time() - _t}
        if cols_to_profile or not is_cached or flush:

            if flush:
                meta = Meta.set(meta, "profile", {})
                df.meta = meta

            hist_cols = []
            freq_cols = []

            cols_data_types = {}
            cols_to_infer = [*cols_to_profile]

            for col_name in cols_to_profile:
                col_data_type = Meta.get(df.meta, f"columns_data_types.{col_name}")

                if col_data_type is not None:
                    cols_data_types[col_name] = col_data_type
                    cols_to_infer.remove(col_name)

            if cols_to_infer:
                cols_data_types = {**cols_data_types, **df.cols.infer_type(cols_to_infer, tidy=False)["infer_type"]}
                cols_data_types = {col: cols_data_types[col] for col in cols_to_profile if col in cols_data_types}

            _t = time.process_time()
            mismatch = df.cols.quality(cols_data_types)
            profiler_time["count_mismatch"] = {
                "columns": cols_data_types, "elapsed_time": time.process_time() - _t}

            # Get with columns are numerical and does not have mismatch so we can calculate the histogram
            cols_properties = cols_data_types.items()
            for col_name, properties in cols_properties:
                if (properties.get("data_type") in df.constants.NUMERIC_TYPES
                    and not properties.get("categorical", False)) \
                        or col_name in force_hist:
                    hist_cols.append(col_name)
                else:
                    freq_cols.append(col_name)

            hist = None
            freq = {}
            sliced_freq = {}
            count_uniques = None
            if len(hist_cols) and bins > 0:
                _t = time.process_time()
                hist = df.cols.hist(hist_cols, buckets=bins, compute=False)
                profiler_time["hist"] = {
                    "columns": hist_cols, "elapsed_time": time.process_time() - _t}

            if len(freq_cols) and bins > 0:
                _t = time.process_time()
                sliced_cols = []
                non_sliced_cols = []

                # Extract the columns with cells larger thatn
                max_cell_length = getattr(df.meta, "max_cell_length", None)

                if max_cell_length:
                    for i, j in max_cell_length.items():
                        if i in freq_cols:
                            if j > 50:
                                sliced_cols.append(i)
                            else:
                                non_sliced_cols.append(i)

                else:
                    non_sliced_cols = freq_cols

                if len(non_sliced_cols) > 0:
                    # print("non_sliced_cols",non_sliced_cols)
                    freq = df.cols.frequency(
                        non_sliced_cols, n=bins, count_uniques=True, compute=False)

                if len(sliced_cols) > 0:
                    # print("sliced_cols", sliced_cols)
                    sliced_freq = df.cols.slice(sliced_cols, 0, 50).cols.frequency(sliced_cols, n=bins,
                                                                                   count_uniques=True,
                                                                                   compute=False)

                profiler_time["frequency"] = {
                    "columns": freq_cols, "elapsed_time": time.process_time() - _t}

            def merge(_columns, _hist, _freq, _mismatch, _data_types, _count_uniques):
                _c = {}

                _hist = {} if _hist is None else _hist.get("hist", {})
                _freq = {} if _freq is None else _freq.get("frequency", {})

                for _col_name in _columns:
                    _c[_col_name] = {
                        "stats": _mismatch.get(_col_name, None),
                        "data_type": _data_types.get(_col_name, None)
                    }
                    if _col_name in _freq:
                        f = _freq[_col_name]
                        _c[_col_name]["stats"]["frequency"] = f["values"]
                        _c[_col_name]["stats"]["count_uniques"] = f["count_uniques"]

                    elif _col_name in _hist:
                        h = _hist[_col_name]
                        _c[_col_name]["stats"]["hist"] = h

                return {"columns": _c}

            # TO-DO: this should be always 0?
            total_count_na = 0

            data_types = df.cols.data_type("*", tidy=False)["data_type"]

            hist, freq, sliced_freq, mismatch = df.functions.compute(
                hist, freq, sliced_freq, mismatch)

            freq = {**freq, **sliced_freq}

            updated_columns = merge(
                cols_to_profile, hist, freq, mismatch, data_types, count_uniques)

            for col in list(profiler_data.get("columns", {}).keys()):
                if col in updated_columns["columns"]:
                    del profiler_data["columns"][col]

            profiler_data = update_dict(profiler_data, updated_columns)

            assign(profiler_data, "name", Meta.get(df.meta, "name"), dict)
            assign(profiler_data, "file_name",
                   Meta.get(df.meta, "file_name"), dict)

            summary = {
                'rows_count': df.rows.count(),
                'missing_count': total_count_na
            }

            if size is True:
                summary.update({'size': df.size(format="human")})

            summary.update(df.profile._columns_summary())

            assign(profiler_data, "summary", summary, dict)

            assign(profiler_data, "summary.missing_count", total_count_na, dict)

            rows_count = df.rows.count()

            if rows_count:
                assign(profiler_data, "summary.p_missing", round(
                    total_count_na / rows_count * 100, 2))
            else:
                assign(profiler_data, "summary.p_missing", None)

        # _t = time.process_time()

        all_columns_names = df.cols.names()

        # meta = Meta.set(meta, "transformations", value={})

        # Order columns
        actual_columns = profiler_data["columns"]
        profiler_data["columns"] = {key: actual_columns[key]
                                    for key in all_columns_names if key in actual_columns}
        meta = Meta.set(meta, "profile", profiler_data)

        if cols_data_types is not None:
            df.meta = meta
            df = df.cols.set_data_type(cols_data_types, inferred=True)
            meta = df.meta

        # Reset Actions
        meta = Meta.reset_actions(meta, parse_columns(df, cols or []))
        df.meta = meta
        profiler_time["end"] = {"elapsed_time": time.process_time() - _t}
        # print(profiler_time)
        return df



from distributed import Client

from optimus.engines.base.dask.constants import STARTING_DASK
from optimus.helpers.constants import JUST_CHECKING
from optimus.helpers.logger import logger


class Dask:
    def __init__(self):
        self._dask = None
        self._client = None

    def create(self, *args, **kwargs):
        logger.print(JUST_CHECKING)
        logger.print("-----")

        logger.print(STARTING_DASK)

        # Create Dask client
        self._client = Client(*args, **kwargs)

        # Print cluster info
        # self._client.scheduler_info()["workers]

        return self

    def load(self, session):
        self._client = session
        return self

    @property
    def dask(self):
        return self._client

    def compute(self, *args, **kwargs):
        return self._client.compute(*args, **kwargs)



import uuid

from optimus.server.functions import run, code, engine, session, create_session, delete_session, features

from .engine import Engine

defaultEngineConfiguration = { "engine": "dask" }

class Session:
    
    def __init__(self, config=None, id=None, **kwargs):
        self.config = config or kwargs
        self.id = id or str(uuid.uuid4())
        self._default_engine = None
        if "engineConfiguration" not in self.config:
            self.config.update({"engineConfiguration": defaultEngineConfiguration})

    def engine(self, config=None, **kwargs):
        config = config or kwargs or self.config["engineConfiguration"]
        e = Engine(config, session=self)
        if self._default_engine is None:
            self._default_engine = e

        return e

    @property
    def default_engine(self):
        if self._default_engine is None:
            return self.engine()

        return self._default_engine

    def request(self, body=None, **kwargs):
        return self.default_engine.request(body, **kwargs)
        



import os
import platform
import sys
from shutil import rmtree

from deepdiff import DeepDiff
from optimus._version import __version__
from pyspark.sql import DataFrame

import optimus.helpers.functions_spark
from optimus.engines.base.engine import BaseEngine
from optimus.engines.spark.create import Create
from optimus.engines.spark.io.jdbc import JDBC
from optimus.engines.spark.io.load import Load
from optimus.engines.spark.spark import Spark
from optimus.helpers.constants import *
from optimus.helpers.core import val_to_list
from optimus.helpers.functions_spark import append as append_df
from optimus.helpers.logger import logger
from optimus.helpers.output import output_json
from optimus.helpers.raiseit import RaiseIt
from optimus.optimus import Engine, EnginePretty

# Singletons
Spark.instance = None


class SparkEngine(BaseEngine):
    __version__ = __version__

    def __init__(self, session=None, master="local[*]", app_name="optimus", checkpoint=False, path=None,
                 file_system="local",
                 verbose=False,
                 repositories=None,
                 packages=None,
                 jars=None,
                 driver_class_path=None,
                 options=None,
                 additional_options=None,
                 load_avro=False,
                 ):

        """
        Transform and roll out
        :param master: 'Master', 'local' or ip address to a cluster
        :param app_name: Spark app name
        :param path: path to the checkpoint folder
        :param checkpoint: If True create a checkpoint folder
        :param file_system: 'local' or 'hadoop'
        :param additional_options:


        :param options: Configuration options that are passed to spark-submit.
            See `the list of possible options
            <https://spark.apache.org/docs/2.4.1/configuration.html#available-properties>`_.
            Note that any options set already through PYSPARK_SUBMIT_ARGS will override
            these.
        :type options: (dict[str,str])
        :param repositories: List of additional maven repositories for package lookup.
        :type repositories: (list[str])

        :param packages: Spark packages that should be installed.
        :type packages: (list[str])

        :param jars: Full paths to jar files that we want to include to the session.
        :type jars: (list[str])

        """
        self.preserve = False

        if jars is None:
            jars = []

        if driver_class_path is None:
            driver_class_path = []

        if session is None:
            # Creating Spark Session
            # If a Spark session in not passed by argument create one

            self.master = master
            self.app_name = app_name

            if options is None:
                options = {}

            self.options = options

            # Initialize as lists
            self.packages = val_to_list(packages)
            self.repositories = val_to_list(repositories)
            self.jars = val_to_list(jars)
            self.driver_class_path = val_to_list(driver_class_path)

            self.additional_options = additional_options

            self.verbose(verbose)

            # Because avro depends of a external package you can decide if should be loaded
            if load_avro == "2.4":
                self._add_spark_packages(["org.apache.spark:spark-avro_2.12:2.4.3"])

            elif load_avro == "2.3":
                self._add_spark_packages(["com.databricks:spark-avro_2.11:4.0.0"])

            # jdbc_jars = ["/jars/spark-redis-2.4.1-SNAPSHOT-jar-with-dependencies.jar",
            #              "/jars/RedshiftJDBC42-1.2.16.1027.jar",
            #              "/jars/mysql-connector-java-8.0.16.jar",
            #              "/jars/ojdbc8.jar", "/jars/postgresql-42.2.5.jar", "/jars/presto-jdbc-0.224.jar",
            #              "/jars/spark-cassandra-connector_2.11-2.4.1.jar", "/jars/sqlite-jdbc-3.27.2.1.jar",
            #              "/jars/mssql-jdbc-7.4.1.jre8.jar"]
            #
            # self._add_jars(absolute_path(jdbc_jars, "uri"))
            # self._add_driver_class_path(absolute_path(jdbc_jars, "posix"))

            self._create_session()

            if path is None:
                path = os.getcwd()

            if checkpoint is True:
                self._set_check_point_folder(path, file_system)

        else:
            # If a session is passed by arguments just save the reference
            # logger.print("Spark session")
            Spark.instance = Spark().load(session)

        self.client = Spark.instance._spark

        # Initialize Spark
        logger.print("""
                             ____        __  _
                            / __ \____  / /_(_)___ ___  __  _______
                           / / / / __ \/ __/ / __ `__ \/ / / / ___/
                          / /_/ / /_/ / /_/ / / / / / / /_/ (__  )
                          \____/ .___/\__/_/_/ /_/ /_/\__,_/____/
                              /_/
                              """)
        logger.print(STARTING_OPTIMUS)
        logger.print(SUCCESS)

    @property
    def F(self):
        from optimus.engines.spark.functions import SparkFunctions
        return SparkFunctions(self)

    @property
    def create(self):
        return Create(self)

    @property
    def load(self):
        return Load(self)

    @property
    def engine(self):
        return Engine.SPARK.value

    @property
    def engine_label(self):
        return EnginePretty.SPARK.value

    @staticmethod
    def connect(driver=None, host=None, database=None, user=None, password=None, port=None, schema="public",
                oracle_tns=None, oracle_service_name=None, oracle_sid=None, presto_catalog=None,
                cassandra_keyspace=None, cassandra_table=None):
        """
        Create the JDBC string connection
        :return: JDBC object
        """

        return JDBC(host, database, user, password, port, driver, schema, oracle_tns, oracle_service_name, oracle_sid,
                    presto_catalog, cassandra_keyspace, cassandra_table)

    @staticmethod
    def output(output):
        """
        Change the table output style
        :param output: "ascii" or "html"
        :return:
        """
        DataFrame.output = output

    @property
    def spark(self):
        """
        Return a Spark session object
        :return:
        """
        return Spark.instance.spark

    @property
    def sc(self):
        """
        Return a Spark Context object
        :return:
        """
        return Spark.instance.sc

    @staticmethod
    def stop():
        """
        Stop Spark Session
        :return:
        """
        Spark.instance.spark.stop()

    @staticmethod
    def _set_check_point_folder(path, file_system):
        """
        Function that receives a workspace path where a folder is created.
        This folder will store temporal dataframes when user writes the .checkPoint().

        :param path: Location of the dataset (string).
        :param file_system: Describes if file system is local or hadoop file system.

        """

        print_check_point_config(file_system)

        if file_system == "hadoop":
            folder_path = path + "/" + "checkPointFolder"
            SparkEngine.delete_check_point_folder(path=path, file_system=file_system)

            # Creating file:
            logger.print("Creating the hadoop folder...")
            command = "hadoop fs -mkdir " + folder_path
            logger.print("$" + command)
            os.system(command)
            logger.print("Hadoop folder created. \n")

            logger.print("Setting created folder as checkpoint folder...")
            Spark.instance.sc.setCheckpointDir(folder_path)
        elif file_system == "local":
            # Folder path:
            folder_path = path + "/" + "checkPointFolder"
            # Checking if tempFolder exits:
            logger.print("Deleting previous folder if exists...")
            if os.path.isdir(folder_path):
                # Deletes folder if exits:
                rmtree(folder_path)

            logger.print("Creating the checkpoint directory...")
            # Creates new folder:
            os.mkdir(folder_path)

            Spark.instance.sc.setCheckpointDir(dirName="file:///" + folder_path)
        else:
            RaiseIt.value_error(file_system, ["hadoop", "local"])

    @staticmethod
    def delete_check_point_folder(path, file_system):
        """
        Function that deletes the temporal folder where temp files were stored.
        The path required is the same provided by user in setCheckPointFolder().

        :param path: path where the info will be saved
        :param file_system: Describes if file system is local or hadoop file system.
        :return:
        """

        if file_system == "hadoop":
            # Folder path:
            folder_path = path + "/" + "checkPointFolder"
            logger.print("Deleting checkpoint folder...")
            command = "hadoop fs -rm -r " + folder_path
            os.system(command)
            logger.print("$" + command)
            logger.print("Folder deleted.")
        elif file_system == "local":
            logger.print("Deleting checkpoint folder...")
            # Folder path:
            folder_path = path + "/" + "checkPointFolder"
            # Checking if tempFolder exits:
            if os.path.isdir(folder_path):
                # Deletes folder if exits:
                rmtree(folder_path)
                # Creates new folder:
                logger.print("Folder deleted.")
            else:
                logger.print("Folder deleted.")
        else:
            RaiseIt.value_error(file_system, ["hadoop", "local"])

    @staticmethod
    def append(dfs, like="columns"):
        """
        Append dataframes rows or columns wise
        :param dfs: Dataframes
        :param like: 'columns' or 'rows'
        :return:
        """
        return append_df(dfs, like)

    def _setup_repositories(self):
        if self.repositories:
            return '--repositories {}'.format(','.join(self.repositories))
        else:
            return ''

    # Spark Package
    def _add_spark_packages(self, packages):
        """
        Define the Spark packages that must be loaded at start time
        :param packages:
        :return:
        """

        for p in val_to_list(packages):
            optimus.helpers.functions_spark.append(p)

    def _setup_packages(self):
        if self.packages:
            return '--packages {}'.format(','.join(self.packages))
        else:
            return ''

    # Jar
    def _add_jars(self, jar):
        for j in val_to_list(jar):
            self.jars.append(j)

    def _setup_jars(self):
        if self.jars:
            return '--jars "{}"'.format(','.join(self.jars))
        else:
            return ''

    # Driver class path
    def _add_driver_class_path(self, driver_class_path):

        for d in val_to_list(driver_class_path):
            self.driver_class_path.append(d)

    def _setup_driver_class_path(self):

        p = platform.system()
        logger.print("Operative System:" + p)
        if p == "Windows":
            separator = ";"
        else:  # for p == "Linux" or p == "Darwin" or any other p == value
            separator = ":"

        if self.driver_class_path:
            return '--driver-class-path "{}"'.format(separator.join(self.driver_class_path))
        else:
            return ''

    # Options
    def _setup_options(self, additional_options):
        options = {}

        options.update(self.options)

        if additional_options:
            options.update(additional_options)

        if 'spark.sql.catalogImplementation' not in options:
            options['spark.sql.catalogImplementation'] = 'hive'

        # Here we massage conf properties with the intent to pass them to
        # spark-submit; this is convenient as it is unified with the approach
        # we take for repos, packages and jars, and it also handles precedence
        # of conf properties already defined by the user in a very
        # straightforward way (since we always append to PYSPARK_SUBMIT_ARGS)
        return ' '.join('--conf "{}={}"'.format(*o) for o in sorted(options.items()))

    def _create_session(self):
        """
        Start a Spark session using jar, packages, repositories and options given
        :return:
        """

        # Get python.exe fullpath
        os.environ['PYSPARK_PYTHON'] = sys.executable

        # Remove duplicated strings
        # print(self._setup_jars())
        # print(os.environ.get('PYSPARK_SUBMIT_ARGS', '').replace(self._setup_jars(), ''))

        submit_args = [
            # options that were already defined through PYSPARK_SUBMIT_ARGS
            # take precedence over SparklySession's

            self._setup_repositories(),
            self._setup_packages(),
            self._setup_jars(),
            self._setup_driver_class_path(),
            self._setup_options(self.additional_options),
            'pyspark-shell',
        ]

        if self.preserve:
            submit_args.insert(0, os.environ.get('PYSPARK_SUBMIT_ARGS', '').replace('pyspark-shell', ''))

        env = ' '.join(filter(None, submit_args))
        os.environ['PYSPARK_SUBMIT_ARGS'] = env
        Spark.instance = Spark().create(self.master, self.app_name)

    def has_package(self, package_prefix):
        """
        Check if the package is available in the session.
        :param package_prefix: E.g. "org.elasticsearch:elasticsearch-spark".
        :type package_prefix: str
        :return bool
        """
        return any(package for package in self.packages if package.startswith(package_prefix))

    def has_jar(self, jar_name):
        """
        Check if the jar is available in the session.
        :param jar_name: E.g. "mysql-connector-java".
        :type jar_name: str
        :return: bool
        """
        return any(jar for jar in self.jars if jar_name in jar)

    @staticmethod
    def verbose(verbose):
        """
        Enable verbose mode
        :param verbose:
        :return:
        """

        logger.active(verbose)

    @staticmethod
    def compare(df1, df2, method="json"):
        """
        Compare 2 Spark dataframes
        :param df1:
        :param df2:
        :param method: json or a
        :return:
        """
        if method == "json":
            diff = DeepDiff(df1.to_json(), df2.to_json(), ignore_order=False)
            print(output_json(diff))
        elif method == "collect":
            if df1.collect() == df2.collect():
                print("Dataframes are equal")
                return True
            else:
                print("Dataframes not equal. Use 'json' param to check for differences")
                return False

        else:
            RaiseIt.type_error(method, ["json", "collect"])



import functools
import operator
from optimus.engines.base.distributed.rows import DistributedBaseRows

from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions


class DaskBaseRows(DistributedBaseRows):

    def limit(self, count):
        """
        Limit the number of rows
        :param count:
        :return:
        """
        df = self.root
        # Reference https://stackoverflow.com/questions/49139371/slicing-out-a-few-rows-from-a-dask-dataframe

        if count is None:
            return df

        return self.root.new(
            self.root._base_to_dfd(
                df.cols.select("*").data.head(count, npartitions=-1), df.partitions() if count > 100 else 1
            )
        )

    def _drop_duplicated_builtin(self, cols):

        df = self.root
        dfd = df.data

        dfd = dfd.drop_duplicates(subset=cols).reset_index(drop=True)

        meta = self.root.meta
        meta = Meta.action(meta, Actions.DROP_ROW.value, df.cols.names())

        return self.root.new(dfd, meta=meta)
                
    def drop_duplicated(self, cols="*", keep="first", how="any"):
        # not supported arguments on dask
        if how == "all" and keep == "first":
            df = self.root
            cols = parse_columns(df, cols)
            return self._drop_duplicated_builtin(cols)
        return self._mask(cols, func=self.root.mask.duplicated, drop=True, keep=keep, how=how)

    def between(self, columns, lower_bound=None, upper_bound=None, invert=False, equal=False,
                bounds=None):
        """
        Trim values at input thresholds
        :param upper_bound:
        :param lower_bound:
        :param columns: Columns to be trimmed
        :param invert:
        :param equal:
        :param bounds:
        :return:
        """
        df = self.root
        # TODO: should process string or dates
        # columns = parse_columns(df, columns, filter_by_column_types=df.constants.NUMERIC_TYPES)
        columns = parse_columns(df, columns)
        if bounds is None:
            bounds = [(lower_bound, upper_bound)]

        def _between(_col_name):

            if invert is False and equal is False:
                op1 = operator.gt
                op2 = operator.lt
                opb = operator.__and__

            elif invert is False and equal is True:
                op1 = operator.ge
                op2 = operator.le
                opb = operator.__and__

            elif invert is True and equal is False:
                op1 = operator.lt
                op2 = operator.gt
                opb = operator.__or__

            elif invert is True and equal is True:
                op1 = operator.le
                op2 = operator.ge
                opb = operator.__or__

            sub_query = []
            for bound in bounds:
                _lower_bound, _upper_bound = bound
                sub_query.append(opb(op1(df[_col_name], _lower_bound), op2(df[_col_name], _upper_bound)))
            query = functools.reduce(operator.__or__, sub_query)

            return query

        for col_name in columns:
            df = df.rows.select(_between(col_name))
        meta = Meta.action(df.meta, Actions.DROP_ROW.value, df.cols.names())
        return self.root.new(df.data, meta=meta)

    def approx_count(self):
        """
        Aprox rows count
        :return:
        """
        df = self.root
        return df.rows.count()

    def between_index(self, lower_bound=None, upper_bound=None, cols="*"):
        """

        :param columns:
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        df = self.root
        dfd = df.data

        if lower_bound is not None:
            length = len(dfd)
            dfd = dfd.tail(length - lower_bound, compute=False)

            if upper_bound is not None:
                upper_bound -= lower_bound

        if upper_bound is not None:
            dfd = dfd.head(upper_bound, compute=False)
        
        if lower_bound is not None or upper_bound is not None:
            dfd = dfd.reset_index(drop=True)

        cols = parse_columns(df, cols)

        return self.root.new(dfd[cols])



from time import time

import joblib
import matplotlib.pyplot as plt
import numpy as np
from sklearn import decomposition
from sklearn import metrics
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import KFold
from sklearn.model_selection import cross_validate
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from optimus.engines.base.ml.models import BaseML
from optimus.helpers.columns import parse_columns
from optimus.helpers.converter import format_dict
from optimus.helpers.core import val_to_list
from optimus.helpers.types import *
from optimus.infer import is_numeric


class Model:
    def __init__(self, model, X_train=None, y_train=None, X_test=None, y_test=None, op=None, **kwargs):
        """

        :param model: ML Model
        :param X_train: Training data
        :param y_train: Training labels
        :param X_test: Test data
        :param y_test: Test labels
        :param kwargs:
        """
        self.op = op
        self.model = model
        self.X_train = X_train
        self.y_train = y_train

        self.X_test = X_test
        self.y_test = y_test
        self.y_pred = None
        self.score = None
        self.evaluation = None
        self.kwargs = kwargs

    def predict(self, value):
        """
        Predict a value
        :param value: Can be a value a list of features inside a list
        :return:
        """
        if is_numeric(value):
            value = [[value]]

        return list(self.model.predict(value))

    def predict_proba(self, df, output_col=None, index=None):
        """
        Predict probabilities for a list of features
        :param df: Dataframe used to predict
        :param output_col: Column name to save the prediction
        :param index: Only get and specific columns from the prediction
        :return:
        """
        prediction = self.model.predict_proba(df.data)

        if index is not None:
            prediction = prediction[:, index:index + 1]

        prediction = self.op.create.dataframe(prediction)
        names = [f"{output_col}_{i}" for i, name in enumerate(prediction.cols.names())]

        return df.cols.concat(prediction.cols.rename(names = names))

    def roc_auc(self):
        from yellowbrick.classifier import ROCAUC
        visualizer = ROCAUC(self.model, classes=["0", "1", "2"])

        visualizer.fit(self.X_train, self.y_train)  # Fit the training data to the visualizer
        visualizer.score(self.X_test, self.y_test)  # Evaluate the model on the test data
        visualizer.show()  # Finalize and show the figure

    def precision_recall(self):
        from yellowbrick.classifier import PrecisionRecallCurve
        viz = PrecisionRecallCurve(
            self.model,
            per_class=False,
            cmap="Set1"
        )
        viz.fit(self.X_train, self.y_train)
        viz.score(self.X_test, self.y_test)
        viz.show()

    def confusion_matrix(self):
        from yellowbrick.classifier import ConfusionMatrix
        cm = ConfusionMatrix(self.model, classes=[0, 1, 2])
        cm.fit(self.X_train, self.y_train)
        cm.score(self.X_test, self.y_test)

        cm.show()

        # y_pred = self.model.predict(self.X_test)
        # return metrics.confusion_matrix(self.y_test, y_pred)

    def coef(self):
        return list(self.model.coef_)

    def intercept(self):
        return self.model.intercept_

    def scores(self):
        return self.score

    def evaluate(self):
        return self.evaluation
        # print("%0.2f accuracy with a standard deviation of %0.2f" % (scores.mean(), scores.std()))

    # def score(self):
    #     return self.model.score(self.X_test, self.y_test)

    def plot(self):
        y_pred = self.model.predict(self.X_test)
        # plt.figure(dpi=100)
        plt.scatter(self.X_test, self.y_test, color='black')
        plt.plot(self.X_test, y_pred, color='blue', linewidth=3)

        plt.xticks(())
        plt.yticks(())

        plt.show()

    def plot_PCA(self):
        final_df = self.root.data
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1)
        ax.set_xlabel('Principal Component 1', fontsize=15)
        ax.set_ylabel('Principal Component 2', fontsize=15)

        ax.set_title('2 component PCA', fontsize=20)
        targets = ['Iris-setosa', 'Iris-versicolor', 'Iris-virginica']
        colors = ['r', 'g', 'b']
        for target, color in zip(targets, colors):
            indicesToKeep = final_df['target'] == target
            ax.scatter(final_df.loc[indicesToKeep, 'principal component 1']
                       , final_df.loc[indicesToKeep, 'principal component 2']
                       , c=color
                       , s=50)
        ax.legend(targets)
        ax.grid()

    def plot_elbow(self, start=4, end=12):
        from yellowbrick.cluster import KElbowVisualizer
        visualizer = KElbowVisualizer(self.model, k=(start, end))

        visualizer.fit(self.X_test)  # Fit the data to the visualizer
        visualizer.show()
        #
        # X =self.X_test
        # distorsions = []
        # for k in range(2, total_clusters):
        #     kmeans = KMeans(n_clusters=total_clusters)
        #     kmeans.fit(X)
        #     distorsions.append(kmeans.inertia_)
        #
        # fig = plt.figure(figsize=(15, 5))
        # plt.plot(range(2, total_clusters), distorsions)
        # plt.grid(True)
        # plt.title('Elbow curve')

    def plot_clusters(self):
        reduced_data = PCA(n_components=2).fit_transform(self.X_test)
        kmeans = self.model
        kmeans.fit(reduced_data)

        # Step size of the mesh. Decrease to increase the quality of the VQ.
        h = .02  # point in the mesh [x_min, x_max]x[y_min, y_max].

        # Plot the decision boundary. For that, we will assign a color to each
        x_min, x_max = reduced_data[:, 0].min() - 1, reduced_data[:, 0].max() + 1
        y_min, y_max = reduced_data[:, 1].min() - 1, reduced_data[:, 1].max() + 1
        xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))

        # Obtain labels for each point in mesh. Use last trained model.
        Z = kmeans.predict(np.c_[xx.ravel(), yy.ravel()])

        # Put the result into a color plot
        Z = Z.reshape(xx.shape)
        plt.figure(1)
        plt.clf()
        plt.imshow(Z, interpolation="nearest",
                   extent=(xx.min(), xx.max(), yy.min(), yy.max()),
                   cmap=plt.cm.Paired, aspect="auto", origin="lower")

        plt.plot(reduced_data[:, 0], reduced_data[:, 1], 'k.', markersize=5)
        # Plot the centroids as a white X
        centroids = kmeans.cluster_centers_
        plt.scatter(centroids[:, 0], centroids[:, 1], marker="x", s=169, linewidths=3,
                    color="w", zorder=10)
        # plt.title("K-means clustering on the digits dataset (PCA-reduced data)\n"
        #           "Centroids are marked with white cross")
        plt.xlim(x_min, x_max)
        plt.ylim(y_min, y_max)
        plt.xticks(())
        plt.yticks(())
        plt.show()

    def save(self, path):
        joblib.dump(self.model, path)


class ML(BaseML):
    def __init__(self, root: 'DataFrameType'):
        super().__init__(root)
        self.root = root

    @staticmethod
    def _split(X, y, **kwargs):
        test_size = kwargs.pop('test_size', 0.2)
        train_size = kwargs.pop('train_size', None)
        random_state = kwargs.pop('random_state', 0)
        # TODO: in case of categorilca data stratify stratify = kwargs.pop('stratify', y)
        stratify = kwargs.pop('stratify', None)
        shuffle = kwargs.pop('shuffle', True)
        return train_test_split(X, y,
                                test_size=test_size,
                                train_size=train_size,
                                random_state=random_state,
                                stratify=stratify,
                                shuffle=shuffle)

    @staticmethod
    def _bench_k_means(kmeans, data, labels):
        # Reference https://scikit-learn.org/stable/auto_examples/cluster/plot_kmeans_digits.html
        """Benchmark to evaluate the KMeans initialization methods.

        Parameters
        ----------
        kmeans : KMeans instance
            A :class:`~sklearn.cluster.KMeans` instance with the initialization
            already set.
        name : str
            Name given to the strategy. It will be used to show the results in a
            table.
        data : ndarray of shape (n_samples, n_features)
            The data to cluster.
        labels : ndarray of shape (n_samples,)
            The labels used to compute the clustering metrics which requires some
            supervision.
        """
        t0 = time()
        estimator = make_pipeline(StandardScaler(), kmeans).fit(data)
        fit_time = time() - t0
        results = {}

        results.update({"inertia": estimator[-1].inertia_})
        # Define the metrics which require only the true labels and estimator
        # labels
        clustering_metrics = [
            metrics.homogeneity_score,
            metrics.completeness_score,
            metrics.v_measure_score,
            metrics.adjusted_rand_score,
            metrics.adjusted_mutual_info_score,
        ]
        results.update({m.__name__: m(labels, estimator[-1].labels_) for m in clustering_metrics})

        # The silhouette score requires the full dataset
        results["silhouette_score"] = metrics.silhouette_score(data, estimator[-1].labels_,
                                                               metric="euclidean", sample_size=300, )

        return results

    def linear_regression(self, features, target, *args, **kwargs):
        """
        Fit a linear regression model. This ensure that the data is ready converting to numeric and dropping NA
        :param features:
        :param target:
        :param args:
        :param kwargs:
        :return:
        """
        df = self.root

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        X = df[features]._to_values()
        y = df[target]._to_values()

        X_train, X_test, y_train, y_test = self._split(X, y, **kwargs)

        fit_intercept = kwargs.pop('fit_intercept', 20)
        normalize = kwargs.pop('normalize', "mse")
        copy_X = kwargs.pop('copy_X', None)
        test_size = kwargs.pop('test_size', 0.2)
        n_jobs = kwargs.pop('n_jobs', 2)

        lm = LinearRegression(fit_intercept=fit_intercept,
                              normalize=normalize,
                              copy_X=copy_X,
                              n_jobs=n_jobs
                              )

        kf = KFold(n_splits=5, shuffle=True, random_state=100)
        scoring = {"neg_mean_absolute_error": "neg_mean_absolute_error",
                   "neg_mean_squared_error": "neg_mean_squared_error",
                   "neg_root_mean_squared_error": "neg_root_mean_squared_error",
                   "r2": "r2"
                   }

        score = cross_validate(lm, X_train, y_train.ravel(), cv=kf, scoring=scoring)

        fm = lm.fit(X_train, y_train.ravel())
        model = Model(fm, X_train, y_train, X_test, y_test)
        model.score = {i: list(score["test_" + i]) for i in scoring.keys()}
        model.evaluation = {"accuracy": lm.score(X, y),
                            "standard deviation": score["test_r2"].std()}

        return model

    def logistic_regression(self, features, target, *args, **kwargs):
        """
        Runs a logistic regression for input (text) DataFrame.
        :return: Model with logistic regression and prediction run.
        """

        df = self.root

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        X = df[features]._to_values()
        y = df[target]._to_values()

        X_train, X_test, y_train, y_test = self._split(X, y, **kwargs)

        fit_intercept = kwargs.pop('fit_intercept', True)
        penalty = kwargs.pop('solver', "l2")
        solver = kwargs.pop('solver', "liblinear")
        # For small datasets, ‘liblinear’ is a good choice, whereas ‘sag’ and ‘saga’ are faster for large ones.

        max_iter = kwargs.pop('max_iter', 100)
        n_jobs = kwargs.pop('n_jobs', 1)

        lm = LogisticRegression(
            multi_class="auto", solver="liblinear",
            fit_intercept=fit_intercept,
            penalty=penalty,
            # solver=solver,
            max_iter=max_iter,
            n_jobs=n_jobs
        )

        kf = KFold(n_splits=5, shuffle=True, random_state=100)
        score = cross_validate(lm, X_train, y_train.ravel(), cv=kf)

        #
        fm = lm.fit(X_train, y_train.ravel())
        model = Model(fm, X_train, y_train, X_test, y_test)
        model.evaluation = {"accuracy": fm.score(X, y),
                            "standard deviation": score["test_score"].mean()}
        #
        return model

    def k_means(self, features, target, n_clusters, methods="k-means++", *args, **kwargs):
        df = self.root

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        X = df[features]._to_values()
        y = df[target]._to_values().ravel()
        methods = val_to_list(methods)
        # methods = ["k-means++", "random"]
        scores = {}
        models = []

        X_train, X_test, y_train, y_test = self._split(X, y, **kwargs)

        for method in methods:
            kmeans = KMeans(init=method, n_clusters=n_clusters, n_init=4,
                            random_state=0)
            models.append(kmeans)
            scores[method] = self._bench_k_means(kmeans=kmeans, data=X, labels=y)

        kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(X_train)
        # model = Model(one_list_to_val(models), X, y)
        model = Model(kmeans, X_train, y_train, X_test, y_test, n_cluster=n_clusters)

        model.score = format_dict(scores)
        # https://stackoverflow.com/questions/19197715/scikit-learn-k-means-elbow-criterion
        # https://stackoverflow.com/questions/41540751/sklearn-kmeans-equivalent-of-elbow-method
        # from sklearn.cluster import KMeans
        # from matplotlib import pyplot as plt
        #
        # X =  # <your_data>
        # distorsions = []
        # for k in range(2, 20):
        #     kmeans = KMeans(n_clusters=k)
        #     kmeans.fit(X)
        #     distorsions.append(kmeans.inertia_)
        #
        # fig = plt.figure(figsize=(15, 5))
        # plt.plot(range(2, 20), distorsions)
        # plt.grid(True)
        # plt.title('Elbow curve')

        #
        # pca = PCA(n_components=n_digits).fit(data)
        # kmeans = KMeans(init=pca.components_, n_clusters=n_digits, n_init=1)
        # self.bench_k_means(kmeans=kmeans, name="PCA-based", data=data, labels=labels)

        return model

    def pca(self, features, target=None, *args, **kwargs):
        import pandas as pd
        n_components = kwargs.pop('n_components', 2)

        df = self.root
        features = parse_columns(df, features)
        if target is not None:
            target = parse_columns(df, target)

        X = df[features]._to_values()
        X = StandardScaler().fit_transform(X)
        pca = decomposition.PCA(n_components=n_components)
        principalComponents = pca.fit_transform(X)

        col_names = ["PCA_" + str(i) for i in range(n_components)]
        principal_df = pd.DataFrame(data=principalComponents, columns=col_names)
        if target:
            finalDf = pd.concat([principal_df, df.data[target]], axis=1)
        else:
            finalDf = principal_df
        return self.root.new(finalDf)

    def random_forest(self, features, target, *args, **kwargs):
        """
        Runs a random forest classifier for input DataFrame.
        :param features: List of columns to select for prediction.
        :param target: Column to predict.
        :return: Model with random forest and prediction run.
        """

        df = self.root

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        # LogisticRegression requires that the Dask Array passed to it has known chunk sizes
        # https://stackoverflow.com/questions/61756328/dusk-ml-logisticregression-throws-this-error-notimplementederror-can-not-add
        X = df[features]._to_values()
        y = df[target]._to_values()

        # Split params
        X_train, X_test, y_train, y_test = self._split(X, y, **kwargs)
        # X_train, X_test, y_train, y_test = ML._split(y, **kwargs)

        # Random Forest
        n_estimators = kwargs.pop('n_estimators', 20)
        criterion = kwargs.pop('criterion', "mse")
        max_depth = kwargs.pop('max_depth', None)
        min_samples_split = kwargs.pop('min_samples_split', 2)
        min_samples_leaf = kwargs.pop('min_samples_leaf', 1)
        min_weight_fraction_leaf = kwargs.pop('min_weight_fraction_leaf', 0.)
        max_features = kwargs.pop('max_features', "auto")
        max_leaf_nodes = kwargs.pop('max_leaf_nodes', None)
        min_impurity_decrease = kwargs.pop('min_impurity_decrease', 0.)
        min_impurity_split = kwargs.pop('min_impurity_split', None)
        bootstrap = kwargs.pop('bootstrap', True)
        oob_score = kwargs.pop('oob_score', False)
        n_jobs = kwargs.pop('n_jobs', None)
        random_state = kwargs.pop('random_state', 0)
        verbose = kwargs.pop('verbose', 0)
        warm_start = kwargs.pop('warm_start', False)
        ccp_alpha = kwargs.pop('ccp_alpha', 0.0)
        max_samples = kwargs.pop('max_samples', None)

        regressor = RandomForestRegressor(n_estimators,
                                          criterion,
                                          max_depth,
                                          min_samples_split,
                                          min_samples_leaf,
                                          min_weight_fraction_leaf,
                                          max_features,
                                          max_leaf_nodes,
                                          min_impurity_decrease,
                                          min_impurity_split,
                                          bootstrap,
                                          oob_score,
                                          n_jobs,
                                          random_state,
                                          verbose,
                                          warm_start,
                                          ccp_alpha,
                                          max_samples)
        fm = regressor.fit(X_train, y_train.ravel())
        model = Model(fm, X_train, y_train, X_test, y_test)
        # y_pred = regressor.predict(X_test

        return model



from pyspark.ml import feature, Pipeline
from pyspark.ml.feature import OneHotEncoder, VectorAssembler, Normalizer, IndexToString
import databricks.koalas as ks
from optimus.engines.base.meta import Meta
from optimus.engines.base.ml.constants import INDEX_TO_STRING
from optimus.engines.base.ml.encoding import BaseEncoding
from optimus.helpers.check import is_spark_dataframe
from optimus.helpers.columns import parse_columns, name_col, get_output_cols
from optimus.helpers.constants import Actions
from optimus.helpers.logger import logger
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_, is_str


class Encoding(BaseEncoding):
    def n_gram(df, input_col, n=2):
        """
        Converts the input array of strings inside of a Spark DF into an array of n-grams.
        :param df: Pyspark dataframe to analyze
        :param input_col: Column to analyzer.
        :param n: number of elements per n-gram >=1.
        :return: Spark DataFrame with n-grams calculated.
        """

        is_spark_dataframe(df)

        tokenizer = feature.Tokenizer().setInputCol(input_col) | feature.StopWordsRemover()
        count = feature.CountVectorizer()
        gram = feature.NGram(n=n) | feature.CountVectorizer()
        tf = tokenizer | (count, gram) | feature.VectorAssembler()
        tfidf = tf | feature.IDF().setOutputCol('features')

        tfidf_model = tfidf.fit(df)
        df_model = tfidf_model.transform(df)
        return df_model, tfidf_model

    def index_to_string(df, input_cols, output_cols=None, columns=None, **kargs):
        """
        Maps a column of indices back to a new column of corresponding string values. The index-string mapping is
        either from the ML attributes of the input column, or from user-supplied labels (which take precedence over
        ML attributes).
        :param df: Dataframe to be transformed.
        :param input_cols: Columns to be indexed.
        :param output_cols: Column where the output is going to be saved.
        :param columns:
        :return: Dataframe with indexed columns.
        """
        df_actual = df

        if columns is None:
            input_cols = parse_columns(df, input_cols)
            if output_cols is None:
                output_cols = [name_col(input_col, INDEX_TO_STRING) for input_col in input_cols]
            output_cols = get_output_cols(input_cols, output_cols)
        else:
            input_cols, output_cols = zip(*columns)

        indexers = [IndexToString(inputCol=input_col, outputCol=output_col, **kargs) for input_col, output_col
                    in zip(list(set(input_cols)), list(set(output_cols)))]
        pipeline = Pipeline(stages=indexers)
        df = pipeline.fit(df).transform(df)

        df.meta = Meta.action(df.meta, df_actual, Actions.INDEX_TO_STRING.value, output_cols)

        return df

    def one_hot_encoder(self, cols="*", prefix=None, drop=True, **kwargs):
        """
        Maps a column of label indices to a column of binary vectors, with at most a single one-value.
        :param cols: Columns to be encoded.
        :param prefix: Prefix of the columns where the output is going to be saved.
        :param drop:
        :return: Dataframe with encoded columns.
        """
        df = self.root

        df = df.new(ks.concat([df.data, ks.get_dummies(df[cols].cols.to_string().data, prefix=prefix)], axis=1))

        if drop:
            df = df.cols.drop(cols)

        return df

    # TODO: Must we use the pipeline version?
    def vector_assembler(df, input_cols, output_col=None):
        """
        Combines a given list of columns into a single vector column.
        :param df: Dataframe to be transformed.
        :param input_cols: Columns to be assembled.
        :param output_col: Column where the output is going to be saved.
        :return: Dataframe with assembled column.
        """

        input_cols = parse_columns(df, input_cols)

        if output_col is None:
            output_col = name_col(input_cols, "vector_assembler")

        assembler = [VectorAssembler(inputCols=input_cols, outputCol=output_col)]

        pipeline = Pipeline(stages=assembler)
        df = pipeline.fit(df).transform(df)

        return df

    def normalizer(df, input_cols, output_col=None, p=2.0):
        """
        Transforms a dataset of Vector rows, normalizing each Vector to have unit norm. It takes parameter p, which
        specifies the p-norm used for normalization. (p=2) by default.
        :param df: Dataframe to be transformed
        :param input_cols: Columns to be normalized.
        :param output_col: Column where the output is going to be saved.
        :param p:  p-norm used for normalization.
        :return: Dataframe with normalized columns.
        """

        # Check if columns argument must be a string or list datat ype:
        if not is_(input_cols, (str, list)):
            RaiseIt.type_error(input_cols, ["str", "list"])

        if is_str(input_cols):
            input_cols = [input_cols]

        if is_(input_cols, (float, int)):
            RaiseIt.type_error(input_cols, ["float", "int"])

        # Try to create a vector
        if len(input_cols) > 1:
            df = df.cols.cast(input_cols, "vector")

        if output_col is None:
            output_col = name_col(input_cols, "normalizer")

        # TODO https://developer.ibm.com/code/2018/04/10/improve-performance-ml-pipelines-wide-dataframes-apache-spark-2-3/
        normal = [Normalizer(inputCol=col_name, outputCol=output_col, p=p) for col_name in
                  list(set(input_cols))]

        pipeline = Pipeline(stages=normal)

        df = pipeline.fit(df).transform(df)

        return df



from enum import Enum

from optimus.helpers.logger import logger


# Python to PySpark reference
#
# type(None): NullType,
# bool: BooleanType,
# int: LongType,
# float: DoubleType,
# str: StringType,
# bytearray: BinaryType,
# decimal.Decimal: DecimalType,
# datetime.date: DateType,
# datetime.datetime: TimestampType,
# datetime.time: TimestampType,
# Profiler


class Actions(Enum):
    """
    Actions that modify a columns/rows.
    """
    # COLUMNS
    INFERRED_DATA_TYPE = "inferred_data_type"
    MATCH = "match"
    LOWER = "lower"
    UPPER = "upper"
    PROPER = "proper"
    PAD = "pad"
    TRIM = "trim"
    REVERSE = "reverse"
    REMOVE_ACCENTS = "remove"
    REMOVE_SPECIAL_CHARS = "remove"
    REMOVE_WHITE_SPACES = "remove"
    LEFT = "left"
    RIGHT = "right"
    MID = "mid"
    REPLACE = "replace"
    REPLACE_REGEX = "replace"
    FILL_NA = "fill_na"
    CAST = "cast"
    IS_NA = "is_na"
    Z_SCORE = "z_score"
    NEST = "nest"
    UNNEST = "unnest"
    SET = "set"
    STRING_TO_INDEX = "string_to_index"
    DATE_FORMAT = "date_format"
    FORMAT_DATE = "format_date"
    INDEX_TO_STRING = "index_to_string"
    MIN_MAX_SCALER = "min_max_scaler"
    MAX_ABS_SCALER = "max_abs_scaler"
    STANDARD_SCALER = "standard_scaler"
    APPLY_COLS = "apply_cols"
    YEARS_BETWEEN = "apply_cols"
    IMPUTE = "impute"
    EXTRACT = "extract"
    ABS = "abs"
    MATH = "math"
    VARIANCE = "variance"
    SLICE = "slice"
    CLIP = "clip"
    DROP = "drop"
    KEEP = "keep"
    CUT = "cut"
    CUM_SUM = "cumsum"
    CUM_PROD = "cumprod"
    CUM_MAX = "cummax"
    CUM_MIN = "cummin"
    TO_FLOAT = "to_float"
    TO_INTEGER = "to_integer"
    TO_NUMERIC = "to_numeric"
    TO_BOOL = "to_bool"
    TO_BOOLEAN = "to_boolean"
    TO_STRING = "to_string"
    YEAR = "year"
    MONTH = "month"
    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"
    SECOND = "second"
    WEEKDAY = "weekday"
    APPEND = "append"
    COPY = "copy"
    RENAME = "rename"
    UNIQUE = "unique"
    INFER = "infer"
    WORD_TOKENIZE = "word_tokenize"
    LENGTH = "length"
    GET = "get"
    ITEM = "item"

    # URL Example	http://search.somedb.com:8080/history?era=darkages
    # scheme	http
    # hostname	search.somedb.com
    # port	    8080
    # origin	http://search.somedb.com:8080
    # path	    /history
    # query	    ?era=darkages
    DOMAIN = "domain"
    TOP_DOMAIN = "top_domain"
    SUB_DOMAIN = "sub_domain"
    URL_SCHEME = "url_scheme"
    URL_PATH = "url_path"
    URL_FILE = "url_file"
    URL_QUERY = "url_query"
    URL_FRAGMENT = "url_fragment"
    HOST = "host"
    PORT = "port"

    EMAIL_DOMAIN = "email_domain"
    EMAIL_USER = "email_user"

    FINGERPRINT = "fingerprint"
    NGRAM_FINGERPRINT = "ngram_fingerprint"
    NGRAMS = "ngrams"
    LEVENSHTEIN = "levenshtein"

    # PHONETIC ENCODING
    SOUNDEX = "soundex"
    METAPHONE = "metaphone"
    DOUBLE_METAPHONE = "double_metaphone"
    MATCH_RATING_CODEX = "match_rating_codex"
    NYSIIS = "nysiis"

    # ROWS
    SELECT_ROW = "select_row"
    DROP_ROW = "drop_row"
    BETWEEN_ROW = "between_drop"
    SORT_ROW = "sort_row"

    @staticmethod
    def list():
        return list(map(lambda c: c.value, Actions))


class ProfilerDataTypesQuality(Enum):
    MISMATCH = 0
    MISSING = 1
    MATCH = 2


class ProfilerDataTypesNumeric(Enum):
    OBJECT = 0
    STRING = 1
    FLOAT = 2
    INT = 3
    BOOL = 4
    BOOLEAN = 5
    DATETIME = 6
    ARRAY = 7
    ZIP_CODE = 8
    GENDER = 9
    IP = 10
    URL = 11
    EMAIL = 12
    CREDIT_CARD_NUMBER = 13
    PHONE_NUMBER = 14
    SOCIAL_SECURITY_NUMBER = 15
    HTTP_CODE = 16
    US_STATE = 17
    CATEGORICAL = 18
    MISSING = 19
    NULL = 20
    LIST = 21
    DICT = 22
    TUPLE = 23
    ADDRESS = 24
    BAN = 25
    UUID = 26
    MAC_ADDRESS = 27
    IPV4_ADDRESS = 28
    IPV6_ADDRESS = 29
    UNKNOWN = 30

    @staticmethod
    def list():
        return list(map(lambda c: c.value, ProfilerDataTypesNumeric))


class ProfilerDataTypes(Enum):
    OBJECT = "object"
    STRING = "str"
    FLOAT = "float"
    INT = "int"
    BOOL = "bool"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    ARRAY = "array"
    ZIP_CODE = "zip_code"
    GENDER = "gender"
    IPV4_ADDRESS = "ipv4"
    IPV6_ADDRESS = "ipv6"
    URL = "url"
    EMAIL = "email"
    CREDIT_CARD_NUMBER = "credit_card_number"
    PHONE_NUMBER = "phone_number"
    ADDRESS = "address"
    UUID = "UUID"
    BAN = "ban"
    SOCIAL_SECURITY_NUMBER = "social_security_number"
    HTTP_CODE = "http_code"
    MAC_ADDRESS = "mac_address"
    DRIVER_LICENSE = "driver_license"
    US_STATE = "us_state"
    CATEGORICAL = "categorical"
    MISSING = "missing"
    NULL = "null"
    LIST = "list"
    DICT = "dict"
    TUPLE = "tuple"
    UNKNOWN = "unknown"

    @staticmethod
    def list():
        return list(map(lambda c: c.value, ProfilerDataTypes))

INDEX_TO_DATA_TYPE_FUNC = {i.value: i.name.lower() for i in ProfilerDataTypesNumeric}

class Schemas(Enum):
    S3 = 's3://'
    GCS = 'gcs://'
    GC = 'gc://'
    HTTP = 'http://'
    HTTPS = 'https://'
    FTP = 'ftp://'
    FILE = 'file://'
    AZ = 'az://'
    ADL = 'adl://'
    ABFS = 'abfs://'

    @staticmethod
    def list():
        return list(map(lambda c: c.value, Schemas))


PROFILER_NUMERIC_DTYPES = [ProfilerDataTypes.INT.value, ProfilerDataTypes.FLOAT.value]
PROFILER_STRING_DTYPES = [ProfilerDataTypes.STRING.value, ProfilerDataTypes.BOOLEAN.value,
                          ProfilerDataTypes.DATETIME.value, ProfilerDataTypes.ARRAY.value,
                          ProfilerDataTypes.OBJECT.value, ProfilerDataTypes.GENDER.value,
                          ProfilerDataTypes.IPV4_ADDRESS.value, ProfilerDataTypes.URL.value,
                          ProfilerDataTypes.EMAIL.value, ProfilerDataTypes.CREDIT_CARD_NUMBER.value,
                          ProfilerDataTypes.ZIP_CODE.value, ProfilerDataTypes.PHONE_NUMBER,
                          ProfilerDataTypes.SOCIAL_SECURITY_NUMBER.value,
                          ProfilerDataTypes.HTTP_CODE.value, ProfilerDataTypes.US_STATE.value]

# Strings and Function Messages
JUST_CHECKING = "Just check that all necessary environments vars are present..."
STARTING_OPTIMUS = "Transform and Roll out..."

SUCCESS = "Optimus successfully imported. Have fun :)."

CONFIDENCE_LEVEL_CONSTANT = [50, .67], [68, .99], [90, 1.64], [95, 1.96], [99, 2.57]


def print_check_point_config(filesystem):
    logger.print(
        "Setting checkpoint folder %s. If you are in a cluster initialize Optimus with master='your_ip' as param",
        filesystem)


# For Google Colab
JAVA_PATH_COLAB = "/usr/lib/jvm/java-8-openjdk-amd64"
RELATIVE_ERROR = 10000

US_STATES_NAMES = ["alabama",
                   "alaska",
                   "american samoa",
                   "arizona",
                   "arkansas",
                   "california",
                   "colorado",
                   "connecticut",
                   "delaware",
                   "district of columbia",
                   "federated states of micronesia",
                   "florida",
                   "georgia",
                   "guam",
                   "hawaii",
                   "idaho",
                   "illinois",
                   "indiana",
                   "iowa",
                   "kansas",
                   "kentucky",
                   "louisiana",
                   "maine",
                   "marshall islands",
                   "maryland",
                   "massachusetts",
                   "michigan",
                   "minnesota",
                   "mississippi",
                   "missouri",
                   "montana",
                   "nebraska",
                   "nevada",
                   "new hampshire",
                   "new jersey",
                   "new mexico",
                   "new york",
                   "north carolina",
                   "north dakota",
                   "northern mariana islands",
                   "ohio",
                   "oklahoma",
                   "oregon",
                   "palau",
                   "pennsylvania",
                   "puerto rico",
                   "rhode island",
                   "south carolina",
                   "south dakota",
                   "tennessee",
                   "texas",
                   "utah",
                   "vermont",
                   "virgin islands",
                   "virginia",
                   "washington",
                   "west virginia",
                   "wisconsin",
                   "wyoming"
                   ]
US_STATES_CODE = [
    "al",
    "ak",
    "as",
    "az",
    "ar",
    "ca",
    "co",
    "ct",
    "de",
    "dc",
    "fm",
    "fl",
    "ga",
    "gu",
    "hi",
    "id",
    "il",
    "in",
    "ia",
    "ks",
    "ky",
    "la",
    "me",
    "mh",
    "md",
    "ma",
    "mi",
    "mn",
    "ms",
    "mo",
    "mt",
    "ne",
    "nv",
    "nh",
    "nj",
    "nm",
    "ny",
    "nc",
    "nd",
    "mp",
    "oh",
    "ok",
    "or",
    "pw",
    "pa",
    "pr",
    "ri",
    "sc",
    "sd",
    "tn",
    "tx",
    "ut",
    "vt",
    "vi",
    "va",
    "wa",
    "wv",
    "wi",
    "wy"
]
CURRENCIES = {"$": "dollar",
              "¢": "cent",
              "£": "point",
              "€": "euro",
              "¥": "yen",
              "₹": "indian rupee",
              "₽": "ruble",
              "元": "yuan",
              "¤": "currency",
              "₠": "euro-currency",
              "₡": "colon",
              "₢": "cruzeiro",
              "₣": "french franc",
              "₤": "lira",
              "₥": "mill",
              "₦": "naira",
              "₧": "peseta",
              "₨": "rupee",
              "₩": "won",
              "₪": "new shequel",
              "₫": "dong",
              "₭": "kip",
              "₮": "tugrik",
              "₯": "drachma",
              "₰": "german penny",
              "₱": "peso",
              "₲": "guarani",
              "₳": "austral",
              "₴": "hryvnia",
              "₵": "cedi",
              "₶": "livre tournois",
              "₸": "tenge",
              "₺": "turkish lira",
              "₼": "manat",
              "৲": "bengali rupee mark",
              "৳": "bengali rupee sign",
              "૱": "gujarati rupee sign",
              "௹": "tamil rupee sign",
              "฿": "thai currency bath",
              "៛": "khmer currency reil",
              "㍐": "square yuan",
              "円": "yen character",
              "圆": "yen/yuan character variant one",
              "圎": "yen/yuan character variant two",
              "圓": "yuan character, in hong kong and taiwan",
              "圜": "yen/yuan character variant three",
              "원": "won character",
              "﷼": "rial sign",
              "＄": "fullwidth dollar sign",
              "￠": "fullwidth cent sign",
              "￡": "fullwidth pound sign",
              "￥": "fullwidth yen sign",
              "￦": "fullwidth won sign"}
PYTHON_SHORT_TYPES = {"string": "string",
                      "str": "string",
                      "integer": "int",
                      "int": "int",
                      "float": "float",
                      "double": "double",
                      "bool": "boolean",
                      "boolean": "boolean",
                      "array": "array",
                      "null": "null"
                      }
PYTHON_TYPES = {"string": str, "int": int, "float": float, "boolean": bool}
PROFILER_COLUMN_TYPES = {"categorical", "numeric", "date", "null", "array", "binary"}
PYTHON_TO_PROFILER = {"string": "categorical", "boolean": "categorical", "int": "numeric", "float": "numeric",
                      "float": "numeric", "date": "date", "array": "array", "binary": "binary", "null": "null"}

PROFILER_CATEGORICAL_DTYPES = [ProfilerDataTypes.BOOL.value,
                               ProfilerDataTypes.BOOLEAN.value, ProfilerDataTypes.ZIP_CODE.value,
                               ProfilerDataTypes.STRING.value, ProfilerDataTypes.HTTP_CODE.value,
                               ProfilerDataTypes.IPV4_ADDRESS.value, ProfilerDataTypes.GENDER.value,
                               ProfilerDataTypes.PHONE_NUMBER.value, ProfilerDataTypes.US_STATE.value,
                               ProfilerDataTypes.SOCIAL_SECURITY_NUMBER.value
                               ]

PYTHON_DATE_TO_FORMAT = {"%a": "www",
                         "%A": "wwww",
                         "%w": "w",
                         "%d": "d",
                         "%m": "M",
                         "%b": "MMM",
                         "%B": "MMMM",
                         "%y": "yy",
                         "%Y": "yyyy",
                         "%H": "H",
                         "%I": "h",
                         "%p": "mi",
                         "%M": "mm",
                         "%S": "ss",
                         "%f": "yy",
                         "%z": "XXX"}

DATE_FORMAT_ITEMS = [("YYYY", "%Y"),
                     ("YYY", "%y"),
                     ("YY", "%y"),
                     ("Y", "%Y"),
                     ("yyyy", "%Y"),
                     ("yyy", "%y"),
                     ("yy", "%y"),
                     ("y", "%y"),
                     ("MMMMM", "%B"),
                     ("MMMM", "%B"),
                     ("MMM", "%b"),
                     ("MM", "%m"),
                     ("M", "%m"),
                     ("mmmm", "%B"),
                     ("mmm", "%b"),
                     ("mm", "%M"),
                     ("m", "%M"),
                     ('DD', '%d'),
                     ('dd', '%d'),
                     ('D', '%d'),
                     ('d', '%d'),
                     ('HH', '%H'),
                     ('hh', '%I'),
                     ('sss', '%f'),
                     ('ss', '%S'),
                     ('SS', '%S'),
                     ('TT', '%p'),
                     ('tt', '%p'),
                     ('WWWW', '%A'),
                     ('WWW', '%a'),
                     ('WW', '%A'),
                     ('W', '%a'),
                     ('wwww', '%A'),
                     ('www', '%a'),
                     ('ww', '%A'),
                     ('w', '%a'),
                     ('XXX', '%z')]

DATE_FORMAT_ITEMS_MONTH = [("YYYY", "%Y"),
                           ("YYY", "%y"),
                           ("YY", "%y"),
                           ("Y", "%Y"),
                           ("yyyy", "%Y"),
                           ("yyy", "%y"),
                           ("yy", "%y"),
                           ("y", "%y"),
                           ("MMMMM", "%B"),
                           ("MMMM", "%B"),
                           ("MMM", "%b"),
                           ("MM", "%m"),
                           ("M", "%m"),
                           ("mmmm", "%B"),
                           ("mmm", "%b"),
                           ("mm", "%m"),
                           ("m", "%m"),
                           ('DD', '%d'),
                           ('dd', '%d'),
                           ('D', '%d'),
                           ('d', '%d'),
                           ('HH', '%H'),
                           ('hh', '%I'),
                           ('sss', '%f'),
                           ('ss', '%S'),
                           ('SS', '%S'),
                           ('TT', '%p'),
                           ('tt', '%p'),
                           ('WWWW', '%A'),
                           ('WWW', '%a'),
                           ('WW', '%A'),
                           ('W', '%a'),
                           ('wwww', '%A'),
                           ('www', '%a'),
                           ('ww', '%A'),
                           ('w', '%a'),
                           ('XXX', '%z')]

CONTRACTIONS = [
    ("a'ight", "alright"),
    ("ain't", "am not"),
    ("amn't", "am not"),
    ("arencha", "are not you"),
    ("aren't", "are not"),
    ("'bout", "about"),
    ("cannot", "can not"),
    ("can't", "cannot"),
    ("cap'n", "captain"),
    ("'cause", "because"),
    ("'cept", "except"),
    ("could've", "could have"),
    ("couldn't", "could not"),
    ("couldn't've", "could not have"),
    ("dammit", "damn it"),
    ("daren't", "dare not"),
    ("daresn't", "dare not"),
    ("dasn't", "dare not"),
    ("didn't", "did not"),
    ("doesn't", "does not"),
    ("don't", "do not "),
    ("dunno", "do not know"),
    ("d'ye", "do you "),
    ("e'en", "even"),
    ("e'er", "ever"),
    ("'em", "them"),
    ("everybody's", "everybody is"),
    ("everyone's", "everyone is"),
    ("fo'c'sle", "forecastle"),
    ("'gainst", "against"),
    ("g'day", "good day"),
    ("gimme", "give me"),
    ("giv'n", "given"),
    ("gonna", "going to"),
    ("gon't", "go not "),
    ("gotta", "got to"),
    ("hadn't", "had not"),
    ("had've", "had have"),
    ("hasn't", "has not"),
    ("haven't", "have not"),
    ("he'd", "he would"),
    ("he'll", "he will"),
    ("helluva", "hell of a"),
    ("he's", "he is"),
    ("here's", "here is"),
    ("he've", "he have"),
    ("how'd", "how would"),
    ("howdy", "how do you do "),
    ("how'll", "how will"),
    ("how're", "how are"),
    ("how's", "how is "),
    ("I'd", "I would"),
    ("I'd've", "I would have"),
    ("I'll", "I shall / I will"),
    ("I'm", "I am"),
    ("I'm'a", "I am about to"),
    ("I'm'o", "I am going to"),
    ("innit", "is it not"),
    ("I've", "I have"),
    ("isn't", "is not"),
    ("it'd", "it would"),
    ("it'll", "it will"),
    ("it's", "it is"),
    ("iunno", "I don't know"),
    ("kinda", "kind of"),
    ("let's", "let us"),
    ("ma'am", "madam"),
    ("mayn't", "may not"),
    ("may've", "may have"),
    ("methinks", "me thinks"),
    ("mightn't", "might not"),
    ("might've", "might have"),
    ("mustn't", "must not"),
    ("mustn't've", "must not have"),
    ("must've", "must have"),
    ("'neath", "beneath"),
    ("needn't", "need not"),
    ("nal", "and all"),
    ("ne'er", "never"),
    ("o'clock", "of the clock"),
    ("o'er", "over"),
    ("ol'", "old"),
    ("oughtn't", "ought not"),
    ("'round", "around"),
    ("'s", "is"),
    ("shalln't", "shall not"),
    ("shan't", "shall not"),
    ("she'd", "she would"),
    ("she'll", "she will"),
    ("she's", "she is"),
    ("should've", "should have"),
    ("shouldn't", "should not"),
    ("shouldn't've", "should not have"),
    ("somebody's", "somebody is"),
    ("someone's", "someone is"),
    ("something's", "something is"),
    ("so're", "so are"),
    ("so's", "so is"),
    ("so've", "so have"),
    ("that'll", "that will"),
    ("that're", "that are"),
    ("that's", "that is"),
    ("that'd", "that would "),
    ("there'd", "there would"),
    ("there'll", "there will"),
    ("there're", "there are"),
    ("there's", "there is"),
    ("these're", "these are"),
    ("these've", "these have"),
    ("they'd", "they would"),
    ("they'll", "they will"),
    ("they're", "they are "),
    ("they've", "they have"),
    ("this's", "this is"),
    ("those're", "those are"),
    ("those've", "those have"),
    ("'thout", "without"),
    ("'til", "until"),
    ("'tis", "it is"),
    ("to've", "to have"),
    ("'twas", "it was"),
    ("'tween", "between"),
    ("'twere", "it were"),
    ("wanna", "want to"),
    ("wasn't", "was not"),
    ("we'd", "we would"),
    ("we'd've", "we would have"),
    ("we'll", "we will"),
    ("we're", "we are"),
    ("we've", "we have"),
    ("weren't", "were not"),
    ("whatcha", "what are you"),
    ("what'd", "what did"),
    ("what'll", "what will"),
    ("what're", "what are"),
    ("what's", "what is"),
    ("what've", "what have"),
    ("when's", "when is"),
    ("where'd", "where did"),
    ("where'll", "where will"),
    ("where're", "where are"),
    ("where's", "where is "),
    ("where've", "where have"),
    ("which'd", "which would"),
    ("which'll", "which will"),
    ("which're", "which are"),
    ("which's", "which is"),
    ("which've", "which have"),
    ("who'd", "who would "),
    ("who'd've", "who would have"),
    ("who'll", "who will"),
    ("who're", "who are"),
    ("who's", "who is "),
    ("who've", "who have"),
    ("why'd", "why did"),
    ("why're", "why are"),
    ("why's", "why is "),
    ("willn't", "will not"),
    ("won't", "will not"),
    ("wonnot", "will not "),
    ("would've", "would have"),
    ("wouldn't", "would not"),
    ("wouldn't've", "would not have"),
    ("y'all", "you all"),
    ("y'all'd've", "you all would have"),
    ("y'all'd'n've", "you all would not have "),
    ("y'all're", "you all are "),
    ("y'at", "you at"),
    ("yes'm", "yes madam"),
    ("yes ma'am", "yes madam"),
    ("yessir", "yes sir"),
    ("you'd", "you would"),
    ("you'll", "you will"),
    ("you're", "you are"),
    ("you've", "you have")
]

UNKNOWN_THRESHOLD = 5



import dask
import dask.dataframe as dd
import pandas as pd

from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.constants import Actions
from optimus.helpers.core import one_list_to_val, val_to_list
from optimus.profiler.functions import fill_missing_var_types

from optimus.engines.base.distributed.columns import DistributedBaseColumns


class DaskBaseColumns(DistributedBaseColumns):

    def exec_agg(self, exprs, compute):
        """
        Execute and aggregation
        :param exprs:
        :param compute:
        :return:
        """

        result = self.F.delayed(self.format_agg)(exprs)
        
        if compute:
            result = self.F.compute(result)

        return result

    def append(self, dfs):
        """

        :param dfs:
        :return:
        """

        dfs = val_to_list(dfs)

        df = self.root
        dfd = dd.concat([df.data.reset_index(drop=True), *[_df.data.reset_index(drop=True) for _df in dfs]], axis=1)
        meta = Meta.action(df.meta, Actions.APPEND.value, df.cols.names())
        return self.root.new(dfd, meta=meta)
        # return dfd

    def qcut(self, cols="*", quantiles=4, handle_invalid="skip"):

        df = self.root.data
        cols = parse_columns(df, cols)
        df[cols] = df[cols].map_partitions(pd.qcut, quantiles)
        return df

    # Date operations
    def to_timestamp(self, cols="*", date_format=None, output_cols=None):
        raise NotImplementedError('Not implemented yet')

    def astype(self, cols="*", output_cols=None, *args, **kwargs):
        raise NotImplementedError('Not implemented yet')

    def nest(self, cols="*", separator="", output_col=None, drop=True, shape="string"):
        """
        Merge multiple columns with the format specified
        :param cols: columns to be nested
        :param separator: char to be used as separator at the concat time
        :param shape: final data type, 'array', 'string' or 'vector'
        :param output_col:
        :return: Dask DataFrame
        """

        df = self.root
        cols = parse_columns(df, cols)
        # output_col = val_to_list(output_col)
        # check_column_numbers(cols, 2)
        if output_col is None:
            output_col = name_col(cols)
            # RaiseIt.type_error(output_col, ["str"])

        # output_col = parse_columns(df, output_col, accepts_missing_cols=True)

        output_ordered_columns = df.cols.names()

        def _nest_string(row):
            v = row[cols[0]].astype(str)
            for i in range(1, len(cols)):
                v = v + separator + row[cols[i]].astype(str)
            return v

        def _nest_array(row):
            # https://stackoverflow.com/questions/43898035/pandas-combine-column-values-into-a-list-in-a-new-column/43898233
            # t['combined'] = t.values.tolist()

            v = row[cols[0]].astype(str)
            for i in range(1, len(cols)):
                v += ", " + row[cols[i]].astype(str)
            return "[" + v + "]"

        if shape == "string":
            kw_columns = {output_col: _nest_string}
        else:
            kw_columns = {output_col: _nest_array}

        dfd = df.cols.assign(kw_columns).data

        if output_col not in output_ordered_columns:
            col_index = output_ordered_columns.index(cols[-1]) + 1
            output_ordered_columns[col_index:col_index] = [output_col]

        meta = Meta.action(df.meta, Actions.NEST.value, list(kw_columns.values()))

        if drop is True:
            for input_col in cols:
                if input_col in output_ordered_columns and input_col != output_col:
                    output_ordered_columns.remove(input_col)

        return self.root.new(dfd, meta).cols.select(output_ordered_columns)
